# built-in
import ast
import builtins
from typing import Iterator, Optional

# external
import astroid

# app
from .._contract import Category
from .._stub import StubsManager
from .common import TOKENS, Extractor, Token, get_full_name, get_name, get_stub, infer
from .contracts import get_contracts


get_exceptions = Extractor()


@get_exceptions.register(*TOKENS.ASSERT)
def handle_assert(expr, **kwargs) -> Optional[Token]:
    token_info = dict(line=expr.lineno, col=expr.col_offset)
    return Token(value=AssertionError, **token_info)


# explicit raise
@get_exceptions.register(*TOKENS.RAISE)
def handle_raise(expr, **kwargs) -> Optional[Token]:
    token_info = dict(line=expr.lineno, col=expr.col_offset)
    name = get_name(expr.exc)
    if not name:
        # raised a value, too tricky
        if not isinstance(expr.exc, TOKENS.CALL):
            return None
        # raised an instance of an exception
        name = get_name(expr.exc.func)
        if not name or name[0].islower():
            return None
    exc = getattr(builtins, name, name)
    token_info['col'] = expr.exc.col_offset
    return Token(value=exc, **token_info)


# division by zero
@get_exceptions.register(*TOKENS.BIN_OP)
def handle_bin_op(expr, **kwargs) -> Optional[Token]:
    token_info = dict(line=expr.lineno, col=expr.col_offset)
    if isinstance(expr.op, ast.Div) or expr.op == '/':
        if isinstance(expr.right, astroid.node_classes.NodeNG):
            guesses = infer(expr=expr.right)
            token_info['col'] = expr.right.col_offset
            for guess in guesses:
                if type(guess) is not astroid.Const:
                    continue
                return Token(value=ZeroDivisionError, **token_info)
        if isinstance(expr.right, ast.Num) and expr.right.n == 0:
            token_info['col'] = expr.right.col_offset
            return Token(value=ZeroDivisionError, **token_info)
    return None


@get_exceptions.register(*TOKENS.CALL)
def handle_call(expr, dive: bool = True, stubs: StubsManager = None) -> Iterator[Token]:
    token_info = dict(line=expr.lineno, col=expr.col_offset)
    # exit()
    name = get_name(expr.func)
    if name and name == 'exit':
        yield Token(value=SystemExit, **token_info)
    # sys.exit()
    if isinstance(expr.func, TOKENS.ATTR):
        name = get_name(expr.func)
        if name and name == 'sys.exit':
            yield Token(value=SystemExit, **token_info)

    stubs_found = False
    if type(expr) is astroid.Call and stubs is not None:
        for token in _exceptions_from_stubs(expr=expr, stubs=stubs):
            stubs_found = True
            yield token

    # Infer function call and check the function body for raises.
    # Do not dive into called function if we already found stubs for it.
    if not stubs_found and dive:
        yield from _exceptions_from_func(expr=expr)


def _exceptions_from_stubs(expr: astroid.Call, stubs: StubsManager) -> Iterator[Token]:
    for value in infer(expr=expr.func):
        if type(value) is not astroid.FunctionDef:
            continue
        module_name, func_name = get_full_name(expr=value)
        stub = get_stub(module_name=module_name, expr=value, stubs=stubs)
        if stub is None:
            continue
        names = stub.get(func=func_name, contract=Category.RAISES)
        for name in names:
            name = getattr(builtins, name, name)
            yield Token(value=name, line=expr.lineno, col=expr.col_offset)


def _exceptions_from_func(expr) -> Iterator[Token]:
    for value in infer(expr.func):
        if type(value) is not astroid.FunctionDef:
            continue

        # recursively infer exceptions from the function body
        for error in get_exceptions(body=value.body, dive=False):
            yield Token(value=error.value, line=expr.lineno, col=expr.col_offset)

        # get explicitly specified exceptions from `@deal.raises`
        if not value.decorators:
            continue
        for category, args in get_contracts(value.decorators.nodes):
            if category != 'raises':
                continue
            for arg in args:
                name = get_name(arg)
                if name is None:
                    continue
                yield Token(value=name, line=expr.lineno, col=expr.col_offset)
    return None



# built-in
import typing
from inspect import signature

# external
import hypothesis
import hypothesis.strategies
import typeguard

# app
from ._decorators import Pre, Raises
from ._types import ArgsKwargsType


class TestCase(typing.NamedTuple):
    """A callable object, wrapper around a function that must be tested.

    When called, calls the wrapped function, suppresses expected exceptions,
    checks the type of the result, and returns it.
    """

    args: typing.Tuple[typing.Any, ...]
    """Positional arguments to be passed in the function"""

    kwargs: typing.Dict[str, typing.Any]
    """Keyword arguments to be passed in the function"""

    func: typing.Callable
    """The function which will be called when the test case is called"""

    exceptions: typing.Tuple[typing.Type[Exception], ...]
    """Exceptions that must be suppressed.
    """

    check_types: bool
    """Check that the result matches return type of the function.
    """

    def __call__(self) -> typing.Any:
        """Calls the given test case returning the called functions result on success or
        Raising an exception on error
        """
        try:
            result = self.func(*self.args, **self.kwargs)
        except self.exceptions:
            return typing.NoReturn  # type: ignore
        self._check_result(result)
        return result

    def _check_result(self, result: typing.Any) -> None:
        if not self.check_types:
            return
        hints = typing.get_type_hints(self.func)
        if 'return' not in hints:
            return
        memo = typeguard._CallMemo(
            func=self.func,
            args=self.args,
            kwargs=self.kwargs,
        )
        typeguard.check_argument_types(memo=memo)
        typeguard.check_type(
            argname='the return value',
            value=result,
            expected_type=hints['return'],
            memo=memo,
        )


def get_excs(func: typing.Any) -> typing.Iterator[typing.Type[Exception]]:
    while True:
        if getattr(func, '__closure__', None):
            for cell in func.__closure__:
                obj = cell.cell_contents
                if isinstance(obj, Raises):
                    yield from obj.exceptions
                elif isinstance(obj, Pre):
                    if isinstance(obj.exception, Exception):
                        yield type(obj.exception)
                    else:
                        yield obj.exception

        if not hasattr(func, '__wrapped__'):
            return
        func = func.__wrapped__


def get_validators(func: typing.Any) -> typing.Iterator[typing.Callable]:
    """Returns pre-condition validators.

    It is used in the process of generating hypothesis strategies
    To let hypothesis more effectively avoid wrong input values.
    """
    while True:
        if getattr(func, '__closure__', None):
            for cell in func.__closure__:
                obj = cell.cell_contents
                if isinstance(obj, Pre):
                    yield obj.validate

        if not hasattr(func, '__wrapped__'):
            return
        func = func.__wrapped__


def get_examples(
    func: typing.Callable,
    kwargs: typing.Dict[str, typing.Any],
    count: int,
) -> typing.List[ArgsKwargsType]:
    kwargs = kwargs.copy()
    for name, value in kwargs.items():
        if isinstance(value, hypothesis.strategies.SearchStrategy):
            continue
        kwargs[name] = hypothesis.strategies.just(value)

    def pass_along_variables(*args, **kwargs) -> ArgsKwargsType:
        return args, kwargs

    pass_along_variables.__signature__ = signature(func)    # type: ignore
    pass_along_variables.__annotations__ = getattr(func, '__annotations__', {})
    strategy = hypothesis.strategies.builds(pass_along_variables, **kwargs)
    validators = list(get_validators(func))
    examples = []

    @hypothesis.given(strategy)
    @hypothesis.settings(
        database=None,
        max_examples=count,
        deadline=None,
        verbosity=hypothesis.Verbosity.quiet,
        phases=(hypothesis.Phase.generate,),
        suppress_health_check=hypothesis.HealthCheck.all(),
    )
    def example_generator(ex: ArgsKwargsType) -> None:
        for validator in validators:
            try:
                validator(*ex[0], **ex[1])
            except Exception:
                hypothesis.reject()
        examples.append(ex)

    example_generator()  # pylint: disable=no-value-for-parameter
    return examples


def cases(
    func: typing.Callable, *,
    count: int = 50,
    kwargs: typing.Dict[str, typing.Any] = None,
    check_types: bool = True,
) -> typing.Iterator[TestCase]:
    """[summary]

    :param func: the function to test. Should be type annotated.
    :type func: typing.Callable
    :param count: how many test cases to generate, defaults to 50.
    :type count: int, optional
    :param kwargs: keyword arguments to pass into the function.
    :type kwargs: typing.Dict[str, typing.Any], optional
    :param check_types: check that the result matches return type of the function.
        Enabled by default.
    :type check_types: bool, optional
    :yield: Emits test cases.
    :rtype: typing.Iterator[TestCase]

    ```pycon
    >>> import deal
    >>> @deal.pre(lambda a, b: b != 0)
    ... def div(a: int, b: int) -> float:
    ...   return a / b
    ...
    >>> cases = deal.cases(div)
    >>> next(cases)
    TestCase(args=(), kwargs=..., func=<function div ...>, exceptions=(<class 'PreContractError'>,))
    >>> case = next(cases)
    >>> case()  # execute the test case
    ...
    ```

    """
    if not kwargs:
        kwargs = {}
    params_generator = get_examples(
        func=func,
        count=count,
        kwargs=kwargs,
    )
    for args, kwargs in params_generator:
        yield TestCase(
            args=args,
            kwargs=kwargs,
            func=func,
            exceptions=tuple(get_excs(func)),
            check_types=check_types,
        )



# external
import pytest

# project
from deal._cli import main


def test_calls_command():
    def cmd(argv):
        assert argv == ['a1', 'a2']
        return 13

    result = main(argv=['cmd', 'a1', 'a2'], commands=dict(cmd=cmd))
    assert result == 13


def test_unknown_command(capsys):
    with pytest.raises(SystemExit):
        main(argv=['unknown', 'a1', 'a2'], commands=dict(cmd=None))
    captured = capsys.readouterr()
    exp = "error: argument command: invalid choice: 'unknown' (choose from 'cmd')"
    assert exp in captured.err



# autopep8: off
from tokenize import untokenize
from deal._source import (
    get_validator_source, _extract_assignment, _extract_lambda_body,
    _extract_decorator_args, _get_tokens, _extract_def_name, processors)


def test_get_validator_source():
    c = lambda x: x > 0  # noqa
    assert get_validator_source(c) == 'x > 0'


def test_get_validator_source_multiline_lambda():
    c = (lambda
        x: x > 0    # noqa: E128
        )           # noqa: E124
    assert get_validator_source(c) == 'x > 0'


def test_get_validator_source_multiple_braces():
    c = ((lambda x: x > 0))
    assert 'x > 0' in get_validator_source(c)


def test_get_validator_source_no_source():
    c = eval('lambda x: x>0')
    assert get_validator_source(c) == ''


def test_extract_assignment_no_assigment():
    tokens = _get_tokens(['aragorn'])
    tokens = _extract_assignment(tokens)
    text = untokenize(tokens)
    assert text == 'aragorn'


def test_extract_lambda_body_no_lambda():
    tokens = _get_tokens(['lambda x'])
    tokens = _extract_lambda_body(tokens)
    text = untokenize(tokens)
    assert text == 'lambda x'


def test_extract_decorator_args_no_call():
    tokens = _get_tokens(['@deal.safe'])
    tokens = _extract_decorator_args(tokens)
    text = untokenize(tokens)
    assert text == ' deal.safe'


def test_extract_class_name():
    tokens = _get_tokens(['class LOL', 'pass'])
    tokens = _extract_def_name(tokens)
    text = untokenize(tokens)
    assert text.lstrip() == 'LOL'


def test_no_tokens():
    assert _get_tokens([]) == []
    for processor in processors:
        assert processor([]) == []


def test_name_token():
    assert _get_tokens([]) == []
    for processor in processors:
        tokens = _get_tokens(['aragorn'])
        tokens = processor(tokens)
        text = untokenize(tokens)
        assert text == 'aragorn'



# built-in
from typing import Iterator, List, Tuple

# external
import astroid

# app
from .common import TOKENS, get_name


SUPPORTED_CONTRACTS = {'deal.pre', 'deal.post', 'deal.raises', 'deal.pure', 'deal.has'}
SUPPORTED_MARKERS = {'deal.pure'}


def get_contracts(decorators: List) -> Iterator[Tuple[str, list]]:
    for contract in decorators:
        if isinstance(contract, TOKENS.ATTR):
            name = get_name(contract)
            if name not in SUPPORTED_MARKERS:
                continue
            yield name.split('.')[-1], []

        if isinstance(contract, TOKENS.CALL):
            if not isinstance(contract.func, TOKENS.ATTR):
                continue
            name = get_name(contract.func)
            if name == 'deal.chain':
                yield from get_contracts(contract.args)
            if name not in SUPPORTED_CONTRACTS:
                continue
            yield name.split('.')[-1], contract.args

        # infer assigned value
        if isinstance(contract, astroid.Name):
            assigments = contract.lookup(contract.name)[1]
            if not assigments:
                continue
            # use only the closest assignment
            expr = assigments[0]
            # can it be not an assignment? IDK
            if not isinstance(expr, astroid.AssignName):  # pragma: no cover
                continue
            expr = expr.parent
            if not isinstance(expr, astroid.Assign):  # pragma: no cover
                continue
            yield from get_contracts([expr.value])



# built-in
from typing import Callable, Tuple, Type, TypeVar

# app
from .._exceptions import ContractError, RaisesContractError
from .._types import ExceptionType
from .base import Base


_CallableType = TypeVar('_CallableType', bound=Callable)


class Raises(Base[_CallableType]):
    exception: ExceptionType = RaisesContractError

    def __init__(
        self,
        *exceptions: Type[Exception],
        message: str = None,
        exception: ExceptionType = None,
    ):
        """
        Step 1. Set allowed exceptions list.
        """
        self.exceptions: Tuple[Type[Exception], ...] = exceptions
        super().__init__(
            validator=None,  # type: ignore
            message=message,
            exception=exception,
        )

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        try:
            return self.function(*args, **kwargs)
        except ContractError:
            raise
        except Exception as exc:
            if not isinstance(exc, self.exceptions):
                raise self.exception from exc
            raise

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        try:
            return await self.function(*args, **kwargs)
        except ContractError:
            raise
        except Exception as exc:
            if not isinstance(exc, self.exceptions):
                raise self.exception from exc
            raise

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        try:
            yield from self.function(*args, **kwargs)
        except ContractError:
            raise
        except Exception as exc:
            if not isinstance(exc, self.exceptions):
                raise self.exception from exc
            raise



# built-in
from textwrap import dedent

# external
import astroid

# project
from deal.linter._extractors import get_pre


def test_check_contract():
    text = """
    @deal.pre(lambda left, right: left > right)
    def inner(left, right):
        return left + right

    def outer():
        inner(1, right=2)   # line 7
        inner(2, 3)         # line 8
        inner(2, right=1)
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func = tree.body[-1]
    messages = [tuple(r) for r in get_pre(body=func.body)]
    assert messages == [(7, 4, '1, right=2', None), (8, 4, '2, 3', None)]


def test_cannot_infer():
    text = """
    @deal.pre(lambda left, right: something_else > left)
    def inner(left, right):
        return left + right

    def outer():
        inner(a, right=1)
        inner(1, right=a)
        inner(1, 2)
        unknown(1, right=2)
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func = tree.body[-1]
    messages = list(get_pre(body=func.body))
    assert messages == []


def test_inferred_something_else():
    text = """
    something_else = 1

    @deal.raises(ValueError)
    def inner(a):
        return a

    def outer():
        any(1)
        inner(1)
        something_else(2)
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func = tree.body[-1]
    messages = list(get_pre(body=func.body))
    assert messages == []



import sys
from pathlib import Path
from deal._cli._memtest import memtest_command
from textwrap import dedent
from io import StringIO


def test_has_side_effect(tmp_path: Path, capsys):
    if 'example' in sys.modules:
        del sys.modules['example']
    text = """
        import deal

        a = []

        @deal.pure
        def func(b: int) -> float:
            a.append({b, b+b})
            return None
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = memtest_command(['--count', '1', str(path)], root=tmp_path, stream=stream)
    assert result == 1

    stream.seek(0)
    captured = stream.read()
    assert '/example.py' in captured
    assert 'running func' in captured
    assert 'func(b=0)' in captured
    assert 'set' in captured
    assert 'x1' in captured


def test_no_side_effects(tmp_path: Path, capsys):
    if 'example' in sys.modules:
        del sys.modules['example']
    text = """
        import deal

        @deal.pure
        def func(b: int) -> float:
            return b+b
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = memtest_command(['--count', '1', str(path)], root=tmp_path, stream=stream)
    assert result == 0

    stream.seek(0)
    captured = stream.read()
    assert '/example.py' in captured
    assert 'running func' in captured
    assert 'func(b=0)' not in captured


def test_no_matching_funcs(tmp_path: Path):
    if 'example' in sys.modules:
        del sys.modules['example']
    text = """
        import deal

        def not_pure1(a: int, b: int) -> float:
            return a / b

        @deal.post(lambda result: result > 0)
        def not_pure2(a: int, b: int) -> float:
            return a / b
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = memtest_command([str(path)], root=tmp_path, stream=stream)
    assert result == 0

    stream.seek(0)
    captured = stream.read()
    assert '/example.py' not in captured



"""
**Deal** is a Python library for [design by contract][wiki] (DbC) programming.
See [documentation][docs] for more details.

[wiki]: https://en.wikipedia.org/wiki/Design_by_contract
[docs]: https://deal.readthedocs.io/index.html
"""

# main package info
__title__ = 'deal'
__version__ = '4.4.0'
__author__ = 'Gram Orsinium'
__license__ = 'MIT'


# app
from ._aliases import chain, ensure, has, inv, post, pre, pure, raises, reason, safe
from ._exceptions import *  # noQA
from ._imports import activate, module_load
from ._schemes import Scheme
from ._state import disable, enable, reset
from ._testing import TestCase, cases


__all__ = [
    'cases',
    'Scheme',
    'TestCase',

    # state
    'disable',
    'enable',
    'reset',

    # decorators
    'chain',
    'ensure',
    'has',
    'inv',
    'post',
    'pre',
    'raises',
    'reason',
    'safe',

    # aliases
    'pure',

    # module level
    'module_load',
    'activate',
]



# built-in
from typing import Callable, TypeVar

# app
from .._exceptions import PreContractError
from .._types import ExceptionType
from .base import Base


_CallableType = TypeVar('_CallableType', bound=Callable)


class Pre(Base[_CallableType]):
    """
    Check contract (validator) before function processing.
    Validate input arguments.
    """
    exception: ExceptionType = PreContractError

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        self.validate(*args, **kwargs)
        return self.function(*args, **kwargs)

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        self.validate(*args, **kwargs)
        return await self.function(*args, **kwargs)

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        self.validate(*args, **kwargs)
        yield from self.function(*args, **kwargs)



# external
import pytest

# project
import deal

# app
from .helpers import run_sync


@pytest.mark.parametrize('markers, expected', [
    (['io'], True),
    (['network'], True),
    (['socket'], True),
    (['network', 'stdout'], True),
    (['import', 'network', 'stdout'], True),
    (['stdout'], False),
    (['stderr'], False),
    (['write'], False),
    (['read'], False),
    (['import'], False),
])
def test_has_network(markers: list, expected: bool):
    assert deal.has(*markers).has_network is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], True),
    (['network'], True),
    (['socket'], True),
    (['network', 'stdout'], True),
    (['import', 'network', 'stdout'], True),
    (['stdout'], True),
    (['stderr'], True),
    (['write'], True),
    (['read'], True),
    (['import'], False),
    (['global'], False),
])
def test_has_io(markers: list, expected: bool):
    assert deal.has(*markers).has_io is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], True),
    (['network'], False),
    (['socket'], False),
    (['network', 'stdout'], True),
    (['import', 'network', 'stdout'], True),
    (['stdout'], True),
    (['print'], True),
    (['stderr'], False),
    (['write'], False),
    (['read'], False),
    (['import'], False),
])
def test_has_stdout(markers: list, expected: bool):
    assert deal.has(*markers).has_stdout is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], True),
    (['network'], False),
    (['socket'], False),
    (['network', 'stderr'], True),
    (['network', 'stdout'], False),
    (['import', 'network', 'stderr'], True),
    (['stderr'], True),
    (['stdout'], False),
    (['print'], False),
    (['write'], False),
    (['read'], False),
    (['import'], False),
])
def test_has_stderr(markers: list, expected: bool):
    assert deal.has(*markers).has_stderr is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], False),
    (['network'], False),
    (['socket'], False),
    (['network', 'stderr'], False),
    (['network', 'stdout'], False),
    (['import', 'network', 'stderr'], True),
    (['stderr'], False),
    (['stdout'], False),
    (['print'], False),
    (['write'], False),
    (['read'], False),
    (['import'], True),
])
def test_has_import(markers: list, expected: bool):
    assert deal.has(*markers).has_import is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], False),
    (['network'], False),
    (['socket'], False),
    (['network', 'stderr'], False),
    (['network', 'stdout'], False),
    (['import', 'global', 'stderr'], True),
    (['stderr'], False),
    (['stdout'], False),
    (['print'], False),
    (['write'], False),
    (['read'], False),
    (['import'], False),
    (['global'], True),
    (['nonlocal'], True),
])
def test_has_global(markers: list, expected: bool):
    assert deal.has(*markers).has_global is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], True),
    (['network'], False),
    (['socket'], False),
    (['network', 'stderr'], False),
    (['network', 'stdout'], False),
    (['import', 'read', 'stderr'], True),
    (['stderr'], False),
    (['stdout'], False),
    (['print'], False),
    (['write'], False),
    (['read'], True),
    (['import'], False),
    (['global'], False),
    (['nonlocal'], False),
])
def test_has_read(markers: list, expected: bool):
    assert deal.has(*markers).has_read is expected


@pytest.mark.parametrize('markers, expected', [
    (['io'], True),
    (['network'], False),
    (['socket'], False),
    (['network', 'stderr'], False),
    (['network', 'stdout'], False),
    (['import', 'write', 'stderr'], True),
    (['stderr'], False),
    (['stdout'], False),
    (['print'], False),
    (['read'], False),
    (['write'], True),
    (['import'], False),
    (['global'], False),
    (['nonlocal'], False),
])
def test_has_write(markers: list, expected: bool):
    assert deal.has(*markers).has_write is expected


def test_decorating_regular_function():
    @deal.has()
    def func(msg):
        if msg:
            print(msg)
        return msg

    assert func('') == ''
    with pytest.raises(deal.SilentContractError):
        func('a')


def test_custom_exception():
    @deal.has(exception=KeyError)
    def func(msg):
        if msg:
            print(msg)
        return msg

    assert func('') == ''
    with pytest.raises(KeyError):
        func('a')


def test_custom_message():
    @deal.has(message='oh hi mark')
    def func(msg):
        if msg:
            print(msg)
        return msg

    assert func('') == ''
    with pytest.raises(deal.SilentContractError, match='oh hi mark'):
        func('a')


def test_custom_exc_and_message():
    @deal.has(exception=KeyError, message='oh hi mark')
    def func(msg):
        if msg:
            print(msg)
        return msg

    assert func('') == ''
    with pytest.raises(KeyError, match='oh hi mark'):
        func('a')


def test_custom_exc_instance_and_message():
    @deal.has(exception=KeyError('oh no'), message='oh hi mark')
    def func(msg):
        if msg:
            print(msg)
        return msg

    assert func('') == ''
    with pytest.raises(KeyError, match='oh no'):
        func('a')


def test_decorating_async_function():
    @deal.has()
    async def func(msg):
        if msg:
            print(msg)
        return msg

    assert run_sync(func('')) == ''
    with pytest.raises(deal.SilentContractError):
        run_sync(func('a'))



# built-in
import ast
from pathlib import Path
from typing import List, NamedTuple

# external
import astroid

# app
from ._contract import Category, Contract
from ._extractors import get_contracts, get_definitions


class Func(NamedTuple):
    name: str
    args: ast.arguments
    body: list
    contracts: List[Contract]

    line: int
    col: int

    @classmethod
    def from_path(cls, path: Path) -> List['Func']:
        text = path.read_text()
        tree = astroid.parse(code=text, path=str(path))
        return cls.from_astroid(tree)

    @classmethod
    def from_text(cls, text: str) -> List['Func']:
        tree = astroid.parse(text)
        return cls.from_astroid(tree)

    @classmethod
    def from_ast(cls, tree: ast.Module) -> List['Func']:
        funcs = []
        definitions = get_definitions(tree=tree)
        for expr in tree.body:
            if not isinstance(expr, ast.FunctionDef):
                continue
            contracts = []
            for category, args in get_contracts(expr.decorator_list):
                contract = Contract(
                    args=args,
                    func_args=expr.args,
                    category=Category(category),
                    context=definitions,
                )
                contracts.append(contract)
            funcs.append(cls(
                name=expr.name,
                args=expr.args,
                body=expr.body,
                contracts=contracts,
                line=expr.lineno,
                col=expr.col_offset,
            ))
        return funcs

    @classmethod
    def from_astroid(cls, tree: astroid.Module) -> List['Func']:
        funcs = []
        definitions = get_definitions(tree=tree)
        for expr in tree.body:
            if not isinstance(expr, astroid.FunctionDef):
                continue

            # make signature
            code = 'def f({}):0'.format(expr.args.as_string())
            func_args = ast.parse(code).body[0].args  # type: ignore

            # collect contracts
            contracts = []
            if expr.decorators:
                for category, args in get_contracts(expr.decorators.nodes):
                    contract = Contract(
                        args=args,
                        func_args=func_args,
                        category=Category(category),
                        context=definitions,
                    )
                    contracts.append(contract)

            funcs.append(cls(
                name=expr.name,
                args=func_args,
                body=expr.body,
                contracts=contracts,
                line=expr.lineno,
                col=expr.col_offset,
            ))
        return funcs

    def __repr__(self) -> str:
        return '{name}({cats})'.format(
            name=type(self).__name__,
            cats=', '.join(contract.category.value for contract in self.contracts),
        )



# built-in
import ast
from contextlib import suppress

# external
import astroid

# app
from .common import Extractor, infer


inner_extractor = Extractor()
UNKNOWN = object()


def get_value(expr):
    if isinstance(expr, ast.AST):
        with suppress(ValueError, SyntaxError):
            return ast.literal_eval(expr)

    if isinstance(expr, astroid.node_classes.NodeNG):
        # AttributeError: 'AsStringVisitor3' object has no attribute 'visit_unknown'
        with suppress(AttributeError):  # pragma: no cover
            renderred = expr.as_string()
            with suppress(ValueError, SyntaxError):
                return ast.literal_eval(renderred)

    handler = inner_extractor.handlers.get(type(expr))
    if handler:
        value = handler(expr=expr)
        if value is not UNKNOWN:
            return value

    # astroid inference
    if hasattr(expr, 'infer'):
        for parent_expr in infer(expr=expr):
            if parent_expr == expr:  # avoid recursion
                continue
            value = get_value(expr=parent_expr)
            if value is not UNKNOWN:
                return value
    return UNKNOWN


@inner_extractor.register(astroid.List, astroid.Set, astroid.Tuple)
def handle_collections(expr):
    result = []
    for element_expr in expr.elts:
        value = get_value(expr=element_expr)
        if value is UNKNOWN:
            return UNKNOWN
        result.append(value)

    if type(expr) is astroid.Tuple:
        return tuple(result)
    if type(expr) is astroid.Set:
        return set(result)
    return result



# external
import pytest

# project
import deal

# app
from .helpers import run_sync


@deal.reason(ValueError, lambda x: x == 1)
def func(x):
    if x == 0:
        raise ZeroDivisionError
    if x == 1:
        raise ValueError
    if x == 2:
        raise ValueError
    return x


def test_reason_just_works():
    assert func(3) == 3


@pytest.mark.parametrize('value, exc', [
    (0, ZeroDivisionError),
    (1, ValueError),
    (2, deal.ReasonContractError),
])
def test_reason_exception(value, exc):
    with pytest.raises(exc):
        func(value)


@pytest.mark.parametrize('value, exc', [
    (0, ZeroDivisionError),
    (1, ValueError),
    (2, deal.ReasonContractError),
])
def test_decorating_async_function(value, exc):
    @deal.reason(ValueError, lambda x: x == 1)
    async def func(x):
        if x == 0:
            raise ZeroDivisionError
        if x == 1:
            raise ValueError
        if x == 2:
            raise ValueError
        return x

    assert run_sync(func(3)) == 3
    with pytest.raises(exc):
        run_sync(func(value))


@pytest.mark.parametrize('value, exc', [
    (0, ZeroDivisionError),
    (1, ValueError),
    (2, deal.ReasonContractError),
])
def test_decorating_generator(value, exc):
    @deal.reason(ValueError, lambda x: x == 1)
    def func(x):
        if x == 0:
            raise ZeroDivisionError
        if x == 1:
            raise ValueError
        if x == 2:
            raise ValueError
        yield x

    assert list(func(3)) == [3]
    with pytest.raises(exc):
        list(func(value))



# built-in
import doctest
import re

# external
import pytest

# project
import deal._aliases
from deal._state import state


rex = re.compile(r'deal\.(\w*ContractError)')


class Checker(doctest.OutputChecker):
    def __init__(self):
        self.diff = []

    def check_output(self, want: str, got: str, optionflags: int) -> bool:
        got = rex.sub(r'\1', got)
        ok = super().check_output(want=want, got=got, optionflags=optionflags)
        if not ok:
            self.diff.append((got, want))
        return ok


finder = doctest.DocTestFinder(exclude_empty=True)


@pytest.mark.parametrize('test', finder.find(deal._aliases))
def test_doctest(test):
    state.color = False
    try:
        runner = doctest.DocTestRunner(checker=Checker())
        runner.run(test)
        result = runner.summarize(verbose=False)
        if result.failed:
            print('Kinda diff:')
            print(*runner._checker.diff, sep='\n')
        assert not result.failed
    finally:
        state.color = True



# built-in
import ast
from textwrap import dedent

# external
import astroid
import pytest

# project
from deal.linter._contract import Category, Contract
from deal.linter._func import Func


TEXT = """
import deal

@deal.raises(ValueError, UnknownError, idk_what_is_it())
@notadeal.raises(KeyError)
def f(x):
    return x
"""


def test_exceptions():
    funcs1 = Func.from_ast(ast.parse(TEXT))
    assert len(funcs1) == 1
    funcs2 = Func.from_astroid(astroid.parse(TEXT))
    assert len(funcs2) == 1
    for func in (funcs1[0], funcs2[0]):
        assert len(func.contracts) == 1
        contract = func.contracts[0]
        assert contract.exceptions == [ValueError, 'UnknownError']


def test_repr():
    c = Contract(category=Category.RAISES, args=[], func_args=None)
    assert repr(c) == 'Contract(raises)'


def test_run():
    text = """
    import deal

    @deal.post(lambda x: x > 0)
    def f(x):
        return x
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    assert len(funcs1) == 1
    funcs2 = Func.from_astroid(astroid.parse(text))
    assert len(funcs2) == 1
    for func in (funcs1[0], funcs2[0]):
        assert len(func.contracts) == 1
        c = func.contracts[0]
        assert c.run(1) is True
        assert c.run(-1) is False


def test_resolve_func():
    text = """
    import deal

    def contract(x):
        return x > 0

    @deal.post(contract)
    def f(x):
        ...
    """
    text = dedent(text).strip()
    funcs = Func.from_astroid(astroid.parse(text))
    assert len(funcs) == 2
    func = funcs[-1]
    assert len(func.contracts) == 1
    c = func.contracts[0]
    assert c.run(1) is True
    assert c.run(-1) is False


def test_resolve_lambda():
    text = """
    import deal

    contract = lambda x: x > 0

    @deal.post(contract)
    def f(x):
        ...
    """
    text = dedent(text).strip()
    funcs = Func.from_astroid(astroid.parse(text))
    assert len(funcs) == 1
    func = funcs[0]
    assert len(func.contracts) == 1
    c = func.contracts[0]
    assert c.run(1) is True
    assert c.run(-1) is False


def test_return_message():
    text = """
    import deal

    @deal.post(lambda x: x > 0 or 'oh no!')
    def f(x):
        return x
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    assert len(funcs1) == 1
    funcs2 = Func.from_astroid(astroid.parse(text))
    assert len(funcs2) == 1
    for func in (funcs1[0], funcs2[0]):
        assert len(func.contracts) == 1
        c = func.contracts[0]
        assert c.run(1) is True
        assert c.run(-1) == 'oh no!'


def test_simplified_signature():
    text = """
    import deal

    @deal.post(lambda _: _.a > _.b)
    def f(a, b):
        return a + b
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    assert len(funcs1) == 1
    funcs2 = Func.from_astroid(astroid.parse(text))
    assert len(funcs2) == 1
    for func in (funcs1[0], funcs2[0]):
        assert len(func.contracts) == 1
        c = func.contracts[0]
        assert c.run(3, 2) is True
        assert c.run(2, 3) is False


@pytest.mark.parametrize('source, deps', [
    ('lambda: ...', set()),
    ('lambda a, b: ...', {'a', 'b'}),
    ('lambda *args, **kwargs: ...', {'args', 'kwargs'}),
    ('lambda a, *, b: ...', {'a', 'b'}),
])
def test_arguments(source: str, deps: set):
    text = """
    import deal

    @deal.post({source})
    def f():
        return 2
    """
    text = text.format(source=source)
    text = dedent(text).strip()

    tree = ast.parse(text)
    print(ast.dump(tree))
    funcs1 = Func.from_ast(tree)

    tree = astroid.parse(text)
    print(tree.repr_tree())
    funcs2 = Func.from_astroid(tree)

    for funcs in (funcs1, funcs2):
        assert len(funcs) == 1
        func = funcs[0]
        assert len(func.contracts) == 1
        c = func.contracts[0]
        assert c.arguments == deps


@pytest.mark.parametrize('source, deps', [
    ('lambda a, b: cd', {'cd'}),
    ('lambda a, b: a+b', set()),
    ('lambda a, b: (a+b)/c', {'c'}),

    ('lambda: re.compile()', {'re'}),
    ('lambda a, b: ab.cd()', {'ab'}),
])
def test_dependencies(source: str, deps: set):
    text = """
    import deal

    @deal.post({source})
    def f(a, b):
        return a + b
    """
    text = text.format(source=source)
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))

    tree = astroid.parse(text)
    print(tree.repr_tree())
    funcs2 = Func.from_astroid(tree)

    for funcs in (funcs1, funcs2):
        assert len(funcs) == 1
        func = funcs[0]
        assert len(func.contracts) == 1
        c = func.contracts[0]
        assert c.dependencies == deps


def test_resolve_and_run_dependencies_func_astroid():
    text = """
    import deal
    CONST = 34

    def contract(a):
        return a == CONST

    @deal.post(contract)
    def f(a):
        return a * 2
    """
    text = dedent(text).strip()
    tree = astroid.parse(text)
    print(tree.repr_tree())
    funcs = Func.from_astroid(tree)
    assert len(funcs) == 2
    func = funcs[-1]
    assert len(func.contracts) == 1
    c = func.contracts[0]

    assert c.run(12) is False
    assert c.run(34) is True


def test_resolve_and_run_dependencies_lambda():
    text = """
    import deal

    CONST = 34

    @deal.post(lambda a: a == CONST)
    def f(a):
        return a * 2
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))

    tree = astroid.parse(text)
    print(tree.repr_tree())
    funcs2 = Func.from_astroid(tree)

    for funcs in (funcs1, funcs2):
        assert len(funcs) == 1
        func = funcs[0]
        assert len(func.contracts) == 1
        c = func.contracts[0]

        assert c.run(12) is False
        assert c.run(34) is True


def test_lazy_import_stdlib():
    text = """
    import deal

    @deal.post(lambda a: re.compile('^abc$').match(a))
    def f(a):
        return a * 2
    """
    text = dedent(text).strip()
    funcs = Func.from_ast(ast.parse(text))
    assert len(funcs) == 1
    func = funcs[0]
    assert len(func.contracts) == 1
    c = func.contracts[0]

    assert c.run('bcd') is False
    assert c.run('abc') is True


def test_unresolvable():
    text = """
    import deal

    @deal.post(lambda a: re.compile(unknown))
    def f(a):
        return a * 2
    """
    text = dedent(text).strip()
    funcs = Func.from_ast(ast.parse(text))
    assert len(funcs) == 1
    func = funcs[0]
    assert len(func.contracts) == 1
    c = func.contracts[0]

    with pytest.raises(NameError):
        c.run('bcd')



# built-in
import typing


ERROR_FORMAT = 'DEAL{code:03d}'


class Error:
    __slots__ = ('row', 'col', 'code', 'text', 'value')

    def __init__(self, *, row: int, col: int, code: int, text: str, value: str = None):
        self.row = row
        self.col = col
        self.code = code
        self.text = text
        self.value = value

    @property
    def full_code(self) -> str:
        return ERROR_FORMAT.format(code=self.code)

    @property
    def message(self) -> str:
        msg = self.full_code + ' ' + self.text
        if self.value:
            msg += ' ({})'.format(self.value)
        return msg

    def __iter__(self) -> typing.Iterator[typing.Union[int, str]]:
        yield self.row
        yield self.col
        yield self.message

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        return '{name}(row={row}, col={col}, code={code})'.format(
            name=type(self).__name__,
            row=self.row,
            col=self.col,
            code=self.code,
        )

    def __hash__(self):
        return hash((self.row, self.col, self.code))



# built-in
import ast
import json
from argparse import ArgumentParser
from pathlib import Path
from textwrap import dedent, indent
from typing import Iterable, Iterator, Sequence, Union

# app
from ..linter import Checker
from .._state import state
from ._common import get_paths
from .._colors import highlight, COLORS, NOCOLORS


TEMPLATE = '  {blue}{row}{end}:{blue}{col}{end} {magenta}{code}{end} {yellow}{text}{end}'
VALUE = ' {magenta}({value}){end}'
POINTER = '{magenta}^{end}'


def get_errors(paths: Iterable[Union[str, Path]]) -> Iterator[dict]:
    for arg in paths:
        for path in get_paths(Path(arg)):
            content = path.read_text()
            checker = Checker(
                filename=str(path),
                tree=ast.parse(content),
            )
            lines = content.split('\n')
            for error in checker.get_errors():
                yield dict(
                    path=str(path),
                    row=error.row,
                    col=error.col,
                    code=error.full_code,
                    text=error.text,
                    value=error.value,
                    content=lines[error.row - 1],
                )


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(prog='python3 -m deal lint')
    parser.add_argument('--json', action='store_true', help='json output')
    parser.add_argument('--nocolor', action='store_true', help='colorless output')
    parser.add_argument('paths', nargs='*', default='.')
    return parser


def lint_command(argv: Sequence[str]) -> int:
    """Run linter against the given files.

    ```bash
    python3 -m deal lint project/
    ```

    Options:

    + `--json`: output violations as [json per line][ndjson].
    + `--nocolor`: output violations in human-friendly format but without colors.
      Useful for running linter on CI.

    Exit code is equal to the found violations count.
    See [linter][linter] documentation for more details.

    [ndjson]: http://ndjson.org/
    [linter]: https://deal.readthedocs.io/basic/linter.html
    """
    parser = get_parser()
    args = parser.parse_args(argv)
    prev = None
    errors = list(get_errors(paths=args.paths))
    colors = COLORS

    # disable colors if disabled globally (NO_COLOR specified)
    if not state.color:
        args.nocolor = True
    # disable colors if `--nocolor` specified
    if args.nocolor:
        colors = NOCOLORS
        state.color = False

    for error in errors:
        if args.json:
            print(json.dumps(error))
            continue

        # print file path
        if error['path'] != prev:
            print('{green}{path}{end}'.format(**colors, **error))
        prev = error['path']

        # print message
        line = TEMPLATE.format(**colors, **error)
        if error['value']:
            line += VALUE.format(**colors, **error)
        print(line)

        # print code line
        pointer = ' ' * error['col'] + POINTER.format(**colors)
        content = error['content']
        if not args.nocolor:
            content = highlight(content)
        content += '\n' + pointer
        content = indent(dedent(content), prefix='    ')
        print(content)
    return len(errors)



# built-in
from argparse import ArgumentParser
from types import MappingProxyType
from typing import Callable, Mapping, Sequence

# app
from ._lint import lint_command
from ._stub import stub_command
from ._test import test_command
from ._memtest import memtest_command


CommandsType = Mapping[str, Callable[[Sequence[str]], int]]
COMMANDS: CommandsType = MappingProxyType(dict(
    lint=lint_command,
    memtest=memtest_command,
    stub=stub_command,
    test=test_command,
))


def main(argv: Sequence[str], *, commands: CommandsType = COMMANDS) -> int:
    parser = ArgumentParser(prog='python3 -m deal')
    parser.add_argument('command', choices=sorted(commands))

    args, unknown_argv = parser.parse_known_args(argv)
    command = commands[args.command]
    return command(unknown_argv)



# built-in
import ast
from contextlib import suppress
from functools import partial
from pathlib import Path
from types import SimpleNamespace
from typing import Callable, Iterator, List, NamedTuple, Optional, Tuple

# external
import astroid

# app
from .._stub import EXTENSION, StubFile, StubsManager


TOKENS = SimpleNamespace(
    ASSERT=(ast.Assert, astroid.Assert),
    ASSIGN=(ast.Assign, astroid.Assign),
    ATTR=(ast.Attribute, astroid.Attribute),
    BIN_OP=(ast.BinOp, astroid.BinOp),
    CALL=(ast.Call, astroid.Call),
    EXPR=(ast.Expr, astroid.Expr),
    FOR=(ast.For, astroid.For),
    FUNC=(ast.FunctionDef, astroid.FunctionDef),
    GLOBAL=(ast.Global, astroid.Global),
    IF=(ast.If, astroid.If),
    NAME=(ast.Name, astroid.Name),
    NONLOCAL=(ast.Nonlocal, astroid.Nonlocal),
    RAISE=(ast.Raise, astroid.Raise),
    RETURN=(ast.Return, astroid.Return),
    UNARY_OP=(ast.UnaryOp, astroid.UnaryOp),
    WITH=(ast.With, astroid.With),
    YIELD=(ast.Yield, astroid.Yield),
)


class Token(NamedTuple):
    line: int
    col: int
    value: Optional[object] = None
    # marker name or error message:
    marker: Optional[str] = None


def traverse(body: List) -> Iterator:
    for expr in body:
        # breaking apart statements
        if isinstance(expr, TOKENS.EXPR):
            yield expr.value
            yield from _travers_expr(expr=expr.value)
            continue
        if isinstance(expr, TOKENS.IF + TOKENS.FOR):
            yield from traverse(body=expr.body)
            yield from traverse(body=expr.orelse)
            continue

        # breaking apart try-except
        if isinstance(expr, (ast.Try, astroid.TryExcept)):
            for handler in expr.handlers:
                yield from traverse(body=handler.body)
            yield from traverse(body=expr.orelse)
        if isinstance(expr, (ast.Try, astroid.TryFinally)):
            yield from traverse(body=expr.finalbody)
            continue

        # extracting things
        if isinstance(expr, TOKENS.WITH):
            yield from traverse(body=expr.body)
        elif isinstance(expr, TOKENS.RETURN + TOKENS.ASSIGN):
            yield expr.value
        yield expr


def _travers_expr(expr):
    if isinstance(expr, TOKENS.CALL):
        for subnode in expr.args:
            yield from traverse(body=[subnode])
        for subnode in (expr.keywords or ()):
            yield from traverse(body=[subnode.value])


def get_name(expr) -> Optional[str]:
    if isinstance(expr, ast.Name):
        return expr.id
    if isinstance(expr, astroid.Name):
        return expr.name

    if isinstance(expr, astroid.Attribute):
        left = get_name(expr.expr)
        if left is None:
            return None
        return left + '.' + expr.attrname
    if isinstance(expr, ast.Attribute):
        left = get_name(expr.value)
        if left is None:
            return None
        return left + '.' + expr.attr

    return None


def get_full_name(expr) -> Tuple[str, str]:
    if expr.parent is None:
        return '', expr.name

    if type(expr.parent) is astroid.Module:
        return expr.parent.qname(), expr.name

    if type(expr.parent) in (astroid.FunctionDef, astroid.UnboundMethod):
        module_name, func_name = get_full_name(expr=expr.parent)
        return module_name, func_name + '.' + expr.name

    if type(expr.parent) is astroid.ClassDef:
        module_name, class_name = get_full_name(expr=expr.parent)
        return module_name, class_name + '.' + expr.name

    path, _, func_name = expr.qname().rpartition('.')
    return path, func_name


def infer(expr) -> Tuple[astroid.node_classes.NodeNG, ...]:
    if not isinstance(expr, astroid.node_classes.NodeNG):
        return tuple()
    with suppress(astroid.exceptions.InferenceError, RecursionError):
        guesses = expr.infer()
        if guesses is astroid.Uninferable:  # pragma: no cover
            return tuple()
        return tuple(g for g in guesses if type(g) is not astroid.Uninferable)
    return tuple()


def get_stub(module_name: Optional[str], expr: astroid.FunctionDef, stubs: StubsManager) -> Optional[StubFile]:
    if not module_name:
        return None
    stub = stubs.get(module_name)
    if stub is not None:
        return stub

    module = _get_module(expr=expr)
    if module is None or module.file is None:
        return None  # pragma: no coverage
    path = Path(module.file).with_suffix(EXTENSION)
    if not path.exists():
        return None
    return stubs.read(path=path)


def _get_module(expr: astroid.node_classes.NodeNG) -> Optional[astroid.Module]:
    if type(expr) is astroid.Module:
        return expr
    if expr.parent is None:
        return None
    return _get_module(expr.parent)


class Extractor:
    __slots__ = ('handlers', )

    def __init__(self):
        self.handlers = dict()

    def _register(self, types: Tuple[type], handler: Callable) -> Callable:
        for tp in types:
            # it's here to have `getattr` to get nodes from `ast` module
            # that are available only in some Python versions.
            if tp is None:
                continue  # pragma: no coverage
            self.handlers[tp] = handler
        return handler

    def register(self, *types):
        return partial(self._register, types)

    def handle(self, expr, **kwargs):
        handler = self.handlers.get(type(expr))
        if not handler:
            return
        token = handler(expr=expr, **kwargs)
        if token is None:
            return
        if type(token) is Token:
            yield token
            return
        yield from token

    def __call__(self, body: List, **kwargs) -> Iterator[Token]:
        for expr in traverse(body=body):
            yield from self.handle(expr=expr, **kwargs)



# built-in
import ast
from textwrap import dedent

# external
import astroid

# project
from deal.linter._func import Func
from deal.linter._rules import (
    CheckAsserts, CheckImports, CheckMarkers, CheckPre, CheckRaises, CheckReturns, rules,
)


def test_error_codes():
    codes = [rule.code for rule in rules]
    assert len(codes) == len(set(codes))


def test_error_messages():
    messages = [rule.message for rule in rules]
    assert len(messages) == len(set(messages))


def test_check_pre():
    checker = CheckPre()
    text = """
    @deal.pre(lambda x: x > 0)
    def example(x):
        return -x

    @deal.raises()
    def caller():
        return example(-3)

    # ignore funcs without contracts
    def caller():
        return example(-3)
    """
    text = dedent(text).strip()
    funcs = Func.from_astroid(astroid.parse(text))
    assert len(funcs) == 3
    actual = []
    for func in funcs:
        actual.extend(tuple(err) for err in checker(func))
    expected = [(7, 11, 'DEAL011 pre contract error (-3)')]
    assert actual == expected


def test_check_returns():
    checker = CheckReturns()
    text = """
    @deal.post(lambda x: x > 0)
    def test(a):
        if a:
            return 1
        else:
            return -1
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(6, 15, 'DEAL012 post contract error (-1)')]
        assert actual == expected


def test_check_returns_with_message():
    checker = CheckReturns()
    text = """
    @deal.post(lambda x: x > 0 or 'oh no!')
    def test(a):
        if a:
            return 1
        else:
            return -1
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(6, 15, 'DEAL012 oh no! (-1)')]
        assert actual == expected


def test_check_returns_ok_unresolved():
    checker = CheckReturns()
    text = """
    @deal.post(unknown)
    def test(a):
        return 1
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = tuple(checker(func))
        assert not actual


def test_check_raises():
    checker = CheckRaises()
    text = """
    @deal.raises(ValueError)
    def test(a):
        raise ValueError
        raise KeyError
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(4, 10, 'DEAL021 raises contract error (KeyError)')]
        assert actual == expected


def test_check_raises_without_allowed():
    checker = CheckRaises()
    text = """
    @deal.raises()
    def test(a):
        raise ValueError
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(3, 10, 'DEAL021 raises contract error (ValueError)')]
        assert actual == expected


def test_check_raises_unknown():
    checker = CheckRaises()
    text = """
    @deal.raises()
    def test(a):
        raise UnknownError
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(3, 10, 'DEAL021 raises contract error (UnknownError)')]
        assert actual == expected


def test_check_raises_inherited():
    checker = CheckRaises()
    text = """
    @deal.raises(LookupError)
    def test(a):
        raise KeyError
        raise ValueError
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(4, 10, 'DEAL021 raises contract error (ValueError)')]
        assert actual == expected


def test_check_prints():
    checker = CheckMarkers()
    text = """
    @deal.pure
    def test(a):
        print(1)
        return 1
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(3, 4, 'DEAL046 missed marker (stdout)')]
        assert actual == expected


def test_check_pure():
    checker = CheckMarkers()
    text = """
    @deal.pure
    @deal.post(lambda x: x)  # skip other contracts
    def test(a):
        global b
        return b
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(4, 4, 'DEAL041 missed marker (global)')]
        assert actual == expected


def test_check_pure_no_returns():
    checker = CheckMarkers()
    text = """
    @deal.pure
    def test(a):
        a + 3
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        assert len(actual) == 1
        expected = 'DEAL043 missed marker (io)'
        assert actual[0][2] == expected


def test_check_has_io():
    checker = CheckMarkers()
    text = """
    @deal.pre(lambda a: len(a) > 2)
    @deal.has('io')
    @deal.post(lambda result: result is not None)
    def test(a):
        import sys
        sys.stdout.write(a)
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        assert len(actual) == 1
        expected = 'DEAL042 missed marker (import)'
        assert actual[0][2] == expected


def test_check_has_custom_markers():
    checker = CheckMarkers()
    text = """
    import deal

    @deal.has('database')
    def inner():
        return 1

    @deal.has()
    def outer():
        return inner()
    """
    text = dedent(text).strip()
    funcs = Func.from_astroid(astroid.parse(text))
    func = funcs[-1]
    actual = [tuple(err) for err in checker(func)]
    assert len(actual) == 1
    expected = 'DEAL040 missed marker (database)'
    assert actual[0][2] == expected


def test_check_has_no_has():
    checker = CheckMarkers()
    text = """
    @deal.pre(lambda a: len(a) > 2)
    @deal.post(lambda result: result is not None)
    @deal.raises(ValueError)
    def test(a):
        import sys
        sys.stdout.write(a)
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        assert actual == []


def test_check_asserts():
    checker = CheckAsserts()
    text = """
    def example(a):
        assert False, "oh no!"
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = [tuple(err) for err in checker(func)]
        expected = [(2, 11, 'DEAL031 assert error (False)')]
        assert actual == expected


def test_skip_asserts_in_tests():
    checker = CheckAsserts()
    text = """
    def test_example(a):
        assert False, "oh no!"
    """
    text = dedent(text).strip()
    funcs1 = Func.from_ast(ast.parse(text))
    funcs2 = Func.from_astroid(astroid.parse(text))
    for func in (funcs1[0], funcs2[0]):
        actual = list(checker(func))
        assert actual == []


def test_check_imports():
    checker = CheckImports()
    text = """
    import deal
    from deal import pre
    from not_a_deal import pre
    from .deal import pre
    """
    text = dedent(text).strip()
    for tree in (ast.parse(text), astroid.parse(text)):
        actual = [tuple(err) for err in checker(tree)]
        expected = [(2, 0, 'DEAL001 ' + CheckImports.message)]
        assert actual == expected



# external
import pytest

# project
import deal

# app
from .helpers import run_sync


def test_return_value_fulfils_contract():
    func = deal.post(lambda x: x > 0)(lambda x: -x)
    assert func(-4) == 4

    with pytest.raises(deal.PostContractError):
        func(4)


def test_decorating_async_function():
    @deal.post(lambda x: x > 0)
    async def func(x):
        return -x

    assert run_sync(func(-2)) == 2
    with pytest.raises(deal.PostContractError):
        run_sync(func(2))


def test_decorating_generator():
    @deal.post(lambda x: x <= 8)
    def double(x):
        yield x
        yield x * 2
        yield x * 4

    assert list(double(2)) == [2, 4, 8]
    with pytest.raises(deal.PostContractError):
        list(double(4))



import pytest
from deal._decorators.base import Base


@pytest.mark.parametrize('args, kwargs, f, expected', [
    # args
    (
        [1], dict(),
        lambda x: 0,
        dict(x=1),
    ),
    # kwargs
    (
        [], dict(x=2),
        lambda x: 0,
        dict(x=2),
    ),
    # args and kwargs
    (
        [1, 2], dict(c=3, d=4),
        lambda a, b, c, d: 0,
        dict(a=1, b=2, c=3, d=4),
    ),
    # *args
    (
        [1, 2], dict(),
        lambda *other: 0,
        dict(other=(1, 2)),
    ),
    # **kwargs
    (
        [], dict(a=1, b=2),
        lambda **other: 0,
        dict(a=1, b=2, other=dict(a=1, b=2)),
    ),
])
def test_args_to_vars(args, kwargs, f, expected):
    actual = Base._args_to_vars(None, args=args, kwargs=kwargs, function=f)
    assert actual == expected



# built-in
from functools import partial, update_wrapper
from types import MethodType
from typing import Callable, TypeVar

# app
from .._exceptions import InvContractError
from .._types import ExceptionType
from .base import Base


_CallableType = TypeVar('_CallableType', bound=Callable)


class InvariantedClass:
    _disable_patching: bool = False

    def _validate(self) -> None:
        """
        Step 5 (1st flow) or Step 4 (2nd flow). Process contract for object.
        """
        # disable methods matching before validation
        self._disable_patching = True
        # validation by Invariant.validate
        self._validate_base(self)
        # enable methods matching after validation
        self._disable_patching = False

    def _patched_method(self, method: Callable, *args, **kwargs):
        """
        Step 4 (1st flow). Call method
        """
        self._validate()
        result = method(*args, **kwargs)
        self._validate()
        return result

    def __getattribute__(self, name: str):
        """
        Step 3 (1st flow). Get method
        """
        attr = super().__getattribute__(name)
        # disable patching for InvariantedClass methods
        if name in ('_patched_method', '_validate', '_validate_base', '_disable_patching'):
            return attr
        # disable patching by flag (if validation in progress)
        if self._disable_patching:
            return attr
        # disable patching for attributes (not methods)
        if not isinstance(attr, MethodType):
            return attr
        # patch
        patched_method = partial(self._patched_method, attr)
        return update_wrapper(patched_method, attr)

    def __setattr__(self, name: str, value):
        """
        Step 3 (2nd flow). Set some attribute
        """
        # set
        super().__setattr__(name, value)
        if name == '_disable_patching':
            return
        # validation only after set
        self._validate()


class Invariant(Base[_CallableType]):
    exception: ExceptionType = InvContractError

    def validate(self, obj) -> None:  # type: ignore
        """
        Step 6. Process contract (validator)
        """

        if hasattr(self.validator, 'is_valid') and hasattr(obj, '__dict__'):
            kwargs = obj.__dict__.copy()
            kwargs.pop('_disable_patching', '')
            self._vaa_validation(**kwargs)
        else:
            self._simple_validation(obj)

    def validate_chain(self, *args, **kwargs) -> None:
        self.validate(*args, **kwargs)
        self.child_validator(*args, **kwargs)

    def __call__(self, _class: type):  # type: ignore
        """
        Step 2. Return wrapped class.
        """
        # patch class parents and add method for validation

        # if already invarianted
        if hasattr(_class, '_validate_base'):
            self.child_validator = _class._validate_base  # type: ignore
            patched_class = type(
                _class.__name__,
                (_class, ),
                {'_validate_base': self.validate_chain},
            )
        # if it's first invariant
        else:
            patched_class = type(
                _class.__name__ + 'Invarianted',
                (InvariantedClass, _class),
                {'_validate_base': self.validate},
            )
        # Magic: _validate_base method use Invariant as self, not _class

        # return update_wrapper(patched_class, _class)
        return patched_class



# built-in
from inspect import getdoc
from typing import get_type_hints

# project
import deal


def get_func():
    @deal.pre(lambda x: x > 0)
    @deal.post(lambda x: x > 0)
    @deal.ensure(lambda *args, **kwargs: True)
    @deal.raises(ValueError)
    @deal.safe
    @deal.safe()
    @deal.pure
    @deal.chain(deal.safe, deal.pure)
    def func(x: int) -> int:
        """docs were before docker
        """
        return x

    return func


def test_preserve_type_annotations():
    """
    IMPORTANT: this checks preserving type annotations in runtime.
    mypy is a static analyser and can produce a different result.
    """
    func = get_func()
    annotations = get_type_hints(func)
    assert set(annotations) == {'x', 'return'}
    assert annotations['x'] in ('int', int)
    assert annotations['return'] in ('int', int)


def test_preserve_docstring():
    func = get_func()
    assert getdoc(func).strip() == 'docs were before docker'



# built-in
import ast
import typing
from pathlib import Path

# external
from astroid import AstroidSyntaxError

# app
from ._error import Error
from ._func import Func
from ._rules import Required, rules
from ._stub import StubsManager


class Checker:
    __slots__ = ('_tree', '_filename', '_stubs')
    name = 'deal'
    _rules = rules

    def __init__(self, tree: ast.Module, file_tokens=None, filename: str = 'stdin'):
        self._tree = tree
        self._filename = filename

        paths = list(StubsManager.default_paths)
        if filename != 'stdin':
            paths.append(Path(filename).absolute().parent)
        self._stubs = StubsManager(paths=paths)

    @property
    def version(self):
        # project
        import deal

        return deal.__version__

    def run(self) -> typing.Iterator[tuple]:
        for error in self.get_errors():
            yield tuple(error) + (type(self),)  # type: ignore

    def get_funcs(self) -> typing.List['Func']:
        if self._filename == 'stdin':
            return Func.from_ast(tree=self._tree)
        try:
            return Func.from_path(path=Path(self._filename))
        except AstroidSyntaxError:
            return Func.from_ast(tree=self._tree)

    def get_errors(self) -> typing.Iterator[Error]:
        reported = set()
        for func in self.get_funcs():
            for rule in self._rules:
                if rule.required != Required.FUNC:
                    continue
                for error in rule(func=func, stubs=self._stubs):
                    hs = hash(error)
                    if hs in reported:
                        continue
                    reported.add(hs)
                    yield error

        for rule in self._rules:
            if rule.required != Required.MODULE:
                continue
            yield from rule(tree=self._tree)



# built-in
from typing import Callable, TypeVar

# app
from .._exceptions import PostContractError
from .._types import ExceptionType
from .base import Base


_CallableType = TypeVar('_CallableType', bound=Callable)


class Post(Base[_CallableType]):
    """
    Check contract (validator) after function processing.
    Validate output result.
    """
    exception: ExceptionType = PostContractError

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        result = self.function(*args, **kwargs)
        self.validate(result)
        return result

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        result = await self.function(*args, **kwargs)
        self.validate(result)
        return result

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        for result in self.function(*args, **kwargs):
            self.validate(result)
            yield result




class cached_property:  # noqa: N801
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value



from deal._mem_test import MemoryTracker


def test_mem_dump_no_diff():
    def f():
        return 123

    tracker = MemoryTracker()
    with tracker:
        f()
    assert not tracker.diff


def test_mem_dump_ignore_locals():
    def f():
        a = 456
        b = a
        return b

    tracker = MemoryTracker()
    with tracker:
        f()
    assert not tracker.diff


def test_mem_dump_side_effect():
    a = []

    def f():
        a.append({12})
        return 123

    tracker = MemoryTracker()
    with tracker:
        f()
    assert dict(tracker.diff) == {'set': 1}



# built-in
import sys
from io import StringIO
from pathlib import Path
from textwrap import dedent

# external
import pytest

# project
import deal
from deal._trace import TraceResult
from deal._testing import TestCase
from deal._cli._test import (
    has_pure_contract, sys_path, test_command as command,
    fast_iterator, format_exception, format_coverage, run_cases,
)
from deal.linter._func import Func


def test_safe_violation(tmp_path: Path, capsys):
    if 'example' in sys.modules:
        del sys.modules['example']
    text = """
        import deal

        @deal.pure
        def func(a: int, b: int) -> float:
            return a / b
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = command(['--count', '1', str(path)], root=tmp_path, stream=stream)
    assert result == 1

    stream.seek(0)
    captured = stream.read()
    assert '/example.py' in captured
    assert 'running func' in captured
    assert 'func(a=0, b=0)' in captured
    assert 'ZeroDivisionError' in captured
    assert 'RaisesContractError' in captured


def test_no_violations(tmp_path: Path):
    if 'example' in sys.modules:
        del sys.modules['example']
    text = """
        import deal

        @deal.pure
        def func(a: int, b: int) -> float:
            return a + b

        def not_pure1(a: int, b: int) -> float:
            return a / b

        @deal.post(lambda result: result > 0)
        def not_pure2(a: int, b: int) -> float:
            return a / b
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = command(['--count', '5', str(path)], root=tmp_path, stream=stream)
    assert result == 0

    stream.seek(0)
    captured = stream.read()
    assert '/example.py' in captured
    assert 'running func' in captured
    assert 'not_pure' not in captured
    assert 'func(' not in captured


def test_no_matching_funcs(tmp_path: Path):
    if 'example' in sys.modules:
        del sys.modules['example']
    text = """
        import deal

        def not_pure1(a: int, b: int) -> float:
            return a / b

        @deal.post(lambda result: result > 0)
        def not_pure2(a: int, b: int) -> float:
            return a / b
    """
    path = (tmp_path / 'example.py')
    path.write_text(dedent(text))
    stream = StringIO()
    result = command(['--count', '5', str(path)], root=tmp_path, stream=stream)
    assert result == 0

    stream.seek(0)
    captured = stream.read()
    assert '/example.py' not in captured


def test_sys_path():
    path = Path('example')
    size = len(sys.path)

    assert sys.path[0] != 'example'
    with sys_path(path):
        assert sys.path[0] == 'example'
    assert sys.path[0] != 'example'
    assert len(sys.path) == size

    with sys_path(path):
        del sys.path[0]
    assert len(sys.path) == size


@pytest.mark.parametrize('source, has', [
    ('@deal.pure \ndef f(): 0', True),
    ('@deal.pure() \ndef f(): 0', True),
    ('@deal.has() \ndef f(): 0', True),
    # ('@deal.has\ndef f(): 0', True),
])
def test_has_pure_contract(source: str, has: bool) -> None:
    funcs = Func.from_text(source)
    assert len(funcs) == 1
    assert has_pure_contract(funcs[0]) is has


def test_fast_iterator():
    seq = [1, 2, 3, 4]
    assert list(fast_iterator(iter(seq))) == seq


def test_print_exception():
    try:
        raise deal.PreContractError
    except deal.PreContractError:
        text = format_exception()
    assert text.startswith('    Traceback (most recent call last):\n')
    assert 'test_test.py' in text
    assert 'PreContractError' in text
    assert text.endswith('\x1b[39;49;00m')


@pytest.mark.parametrize('cov_l, all_l, exp', [
    ({2, 3, 4}, {2, 3, 4}, '    coverage <G>100%<E>'),
    ({2, 4}, {2, 3, 4}, '    coverage <Y>67%<E> (missing 3)'),
    ({2, 5}, {2, 3, 4, 5}, '    coverage <Y>50%<E> (missing 3-4)'),
    (set(), {2, 3, 4, 5}, '    coverage <R>0%<E>'),
])
def test_format_coverage_100(cov_l, all_l, exp):
    fake_colors = dict(
        red='<R>',
        yellow='<Y>',
        green='<G>',
        end='<E>',
    )
    tr = TraceResult(0, 0, covered_lines=cov_l, all_lines=all_l)
    text = format_coverage(tr, colors=fake_colors)
    assert text == exp


def test_run_cases_ok():
    def func():
        return 123

    case = TestCase(
        args=(),
        kwargs={},
        func=func,
        exceptions=(),
        check_types=False,
    )
    cases = [case]
    colors = dict(
        blue='<B>',
        yellow='<Y>',
        end='<E>',
    )
    stream = StringIO()
    ok = run_cases(cases=cases, func_name='fname', stream=stream, colors=colors)
    assert ok
    stream.seek(0)
    captured = stream.read()
    assert captured
    assert captured.split('\n')[0] == '  <B>running fname<E>'


def test_run_cases_bad():
    def func(a, b):
        raise ZeroDivisionError

    case = TestCase(
        args=(1, ),
        kwargs=dict(b=2),
        func=func,
        exceptions=(),
        check_types=False,
    )
    cases = [case]
    colors = dict(
        blue='<B>',
        yellow='<Y>',
        end='<E>',
    )
    stream = StringIO()
    ok = run_cases(cases=cases, func_name='fname', stream=stream, colors=colors)
    assert not ok
    stream.seek(0)
    captured = stream.read()
    assert captured
    assert captured.split('\n')[0] == '  <B>running fname<E>'
    assert 'ZeroDivisionError' in captured
    assert 'Traceback' in captured
    assert '    <Y>fname(1, b=2)<E>' in captured



# built-in
from argparse import ArgumentParser
from pathlib import Path
from typing import List, Sequence

# app
from ..linter import StubsManager, generate_stub
from ._common import get_paths


def stub_command(argv: Sequence[str]) -> int:
    """Generate stub files for the given Python files.

    ```bash
    python3 -m deal stub project/
    ```

    Options:

    + `--iterations`: how many time run stub generation against files.
      Every new iteration uses results from the previous ones, improving the result.
      Default: 1.

    Exit code is 0. See [stubs][stubs] documentation for more details.

    [stubs]: https://deal.readthedocs.io/details/stubs.html
    """
    parser = ArgumentParser(prog='python3 -m deal stub')
    parser.add_argument('--iterations', type=int, default=1)
    parser.add_argument('paths', nargs='+')
    args = parser.parse_args(argv)

    paths: List[Path] = []
    for arg in args.paths:
        for path in get_paths(Path(arg)):
            paths.append(path)

    roots = list(StubsManager.default_paths) + list(set(paths))
    stubs = StubsManager(paths=roots)

    for _ in range(args.iterations):
        for path in paths:
            generate_stub(path=path, stubs=stubs)
    return 0



# built-in
import subprocess
import sys


def test_cli_calling():
    result = subprocess.run([sys.executable, '-m', 'deal', 'lint', __file__])
    assert result.returncode == 0


def test_do_not_import_linter():
    """When deal is imported, it must not trigger importing linter.
    """
    old_modules = sys.modules.copy()
    for name in list(sys.modules):
        if name == 'deal' or name.startswith('deal.'):
            del sys.modules[name]
        if name == 'astroid' or name.startswith('astroid.'):
            del sys.modules[name]

    import deal  # noqa: F401

    for name in list(sys.modules):
        if name.startswith('deal.lint'):
            raise ImportError('oh no, linter was imported')
        if name == 'astroid' or name.startswith('astroid.'):
            raise ImportError('oh no, astroid was imported')

    sys.modules = old_modules



# external
import pytest

# project
import deal


def test_chained_contract_decorator():

    @deal.chain(
        deal.pre(lambda x: x != 1),
        deal.pre(lambda x: x != 2),
    )
    def func(x):
        return x * 4

    func(3)
    func(0)
    with pytest.raises(deal.PreContractError):
        func(1)
    with pytest.raises(deal.PreContractError):
        func(2)



# app
from .ensure import Ensure
from .has import Has
from .inv import Invariant
from .post import Post
from .pre import Pre
from .raises import Raises
from .reason import Reason


__all__ = [
    'Ensure',
    'Has',
    'Invariant',
    'Offline',
    'Post',
    'Pre',
    'Raises',
    'Reason',
    'Silent',
]



# built-in
import ast

# external
import astroid
import pytest

# project
from deal.linter._extractors import get_definitions


@pytest.mark.parametrize('source, names', [
    ('import re', {'re'}),
    ('import typing, types', {'typing', 'types'}),
    ('import typing as types', {'types'}),
    ('from . import hi', set()),
    ('from .something import hi', set()),

    ('from typing import List', {'List'}),
    ('from typing import List, Dict', {'List', 'Dict'}),

    ('ab = 2', {'ab'}),
    ('ab = cd = 23', {'ab', 'cd'}),
    ('ab.cd = 2', set()),
])
def test_extract_defs(source: str, names) -> None:
    tree = ast.parse(source)
    print(ast.dump(tree))
    defs = get_definitions(tree)
    assert set(defs) == names

    module = ast.parse('hello')
    for name, stmt in defs.items():
        module.body[0] = stmt
        print(name, '|>', ast.dump(module))
        compile(module, filename='<ast>', mode='exec')

    tree = astroid.parse(source)
    print(tree.repr_tree())
    defs = get_definitions(tree)
    assert set(defs) == names

    module = ast.parse('hello')
    for name, stmt in defs.items():
        module.body[0] = stmt
        print(name, '|>', ast.dump(module))
        compile(module, filename='<ast>', mode='exec')



# built-in
import socket
import sys
from io import StringIO
from typing import Callable, FrozenSet, Type, TypeVar

# app
from .._exceptions import MarkerError, OfflineContractError, SilentContractError
from .._types import ExceptionType
from .base import Base


_CallableType = TypeVar('_CallableType', bound=Callable)


class PatchedStringIO(StringIO):
    def __init__(self, exception: ExceptionType):
        self.exception = exception

    def write(self, *args, **kwargs):
        raise self.exception


class PatchedSocket:
    def __init__(self, exception: ExceptionType = None):
        self.exception = exception

    def __call__(self, *args, **kwargs):
        raise self.exception


class Has(Base[_CallableType]):
    exception: ExceptionType = MarkerError
    markers: FrozenSet[str]

    def __init__(self, *markers, message: str = None, exception: ExceptionType = None):
        """
        Step 1. Set allowed markers.
        """
        self.markers = frozenset(markers)
        self.message = message
        if exception:
            self.exception = exception

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        self.patch()
        try:
            return self.function(*args, **kwargs)
        finally:
            self.unpatch()

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        self.patch()
        try:
            return await self.function(*args, **kwargs)
        finally:
            self.unpatch()

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        generator = self.function(*args, **kwargs)
        while True:
            self.patch()
            try:
                result = next(generator)
            except StopIteration:
                return
            finally:
                self.unpatch()
            yield result

    # known markers

    @property
    def has_network(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'network' in self.markers:
            return True
        if 'socket' in self.markers:
            return True
        return False

    @property
    def has_io(self) -> bool:
        if self.has_read:
            return True
        if self.has_write:
            return True
        if self.has_stdout:
            return True
        if self.has_stderr:
            return True
        if self.has_network:
            return True
        return False

    @property
    def has_stdout(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'print' in self.markers:
            return True
        if 'stdout' in self.markers:
            return True
        return False

    @property
    def has_stderr(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'stderr' in self.markers:
            return True
        return False

    @property
    def has_import(self) -> bool:
        return 'import' in self.markers

    @property
    def has_global(self) -> bool:
        if 'global' in self.markers:
            return True
        if 'nonlocal' in self.markers:
            return True
        return False

    @property
    def has_read(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'read' in self.markers:
            return True
        return False

    @property
    def has_write(self) -> bool:
        if 'io' in self.markers:
            return True
        if 'write' in self.markers:
            return True
        return False

    # patching

    def patch(self):
        if not self.has_network:
            self.true_socket = socket.socket
            socket.socket = PatchedSocket(
                exception=self._get_exception(OfflineContractError),
            )
        if not self.has_stdout:
            self.true_stdout = sys.stdout
            sys.stdout = PatchedStringIO(
                exception=self._get_exception(SilentContractError),
            )
        if not self.has_stderr:
            self.true_stderr = sys.stderr
            sys.stderr = PatchedStringIO(
                exception=self._get_exception(SilentContractError),
            )

    def unpatch(self):
        if not self.has_network:
            socket.socket = self.true_socket
        if not self.has_stdout:
            sys.stdout = self.true_stdout
        if not self.has_stderr:
            sys.stderr = self.true_stderr

    def _get_exception(self, default: Type[Exception]) -> ExceptionType:
        if self.exception is MarkerError:
            if self.message is None:
                return default
            return default(self.message)

        if self.message is None:
            return self.exception
        if isinstance(self.exception, Exception):
            return self.exception
        return self.exception(self.message)



# built-in
import ast

# external
import astroid

# project
from deal.linter._func import Func


TEXT = """
import deal

@deal.post(lambda x: x > 0)
@deal.raises(ValueError, UnknownError, idk_what_is_it())
@notadeal.raises(KeyError)
@deal.notaraises(KeyError)
@something
@another.deco
@something()
def f(x):
    return x

def h(x):
    return x

@something()
def k(x):
    return x
"""


def test_from_text():
    funcs = Func.from_text(TEXT)
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_ast():
    funcs = Func.from_ast(ast.parse(TEXT))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_astroid():
    funcs = Func.from_astroid(astroid.parse(TEXT))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_repr():
    funcs1 = Func.from_ast(ast.parse(TEXT))
    funcs2 = Func.from_astroid(astroid.parse(TEXT))
    for func in (funcs1[0], funcs2[0]):
        assert repr(func) == 'Func(post, raises)'



# built-in
from typing import Callable, Type, TypeVar

# app
from .._exceptions import ReasonContractError
from .._types import ExceptionType
from .base import Base


_CallableType = TypeVar('_CallableType', bound=Callable)


class Reason(Base[_CallableType]):
    exception: ExceptionType = ReasonContractError

    def __init__(self, event: Type[Exception], validator: Callable, *,
                 message: str = None, exception: ExceptionType = None):
        """
        Step 1. Set allowed exceptions list.
        """
        self.event = event
        super().__init__(
            validator=validator,
            message=message,
            exception=exception,
        )

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        try:
            return self.function(*args, **kwargs)
        except self.event as origin:
            try:
                self.validate(*args, **kwargs)
            except self.exception:
                raise self.exception from origin
            raise

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        try:
            return await self.function(*args, **kwargs)
        except self.event as origin:
            try:
                self.validate(*args, **kwargs)
            except self.exception:
                raise self.exception from origin
            raise

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        try:
            yield from self.function(*args, **kwargs)
        except self.event as origin:
            try:
                self.validate(*args, **kwargs)
            except self.exception:
                raise self.exception from origin
            raise



# built-in
import ast
from textwrap import dedent

# external
import astroid
import pytest

# project
from deal.linter._extractors import get_markers
from deal.linter._stub import StubsManager


@pytest.mark.parametrize('text, expected', [
    ('print(1)', ('stdout', )),
    ('print(1, sep=sys.stderr)', ('stdout', )),
    ('print(1, file=sys.stdout)', ('stdout', )),
    ('print(1, file=sys.stderr)', ('stderr', )),
    ('print(1, file=stream)', ()),

    ('import sys\nsys.stdout.write(1)', ('stdout', )),
    ('import sys\nsys.stderr.write(1)', ('stderr', )),

    ('open("fpath", "w")', ('write', )),
    ('open("fpath", mode="w")', ('write', )),
    ('with open("fpath", "w") as f: ...', ('write', )),

    ('with something: ...', ()),     # uninferrable `with` test
    ('something().anything()', ()),  # complex call, cannot get name

    ('with open("fpath") as f: ...', ('read', )),
    ('with open("fpath", "r") as f: ...', ('read', )),
    ('with open("fpath", mode="r") as f: ...', ('read', )),
    ('open("fpath", "r")', ('read', )),
    ('open("fpath")', ('read', )),
    ('open("fpath", encoding="utf8")', ('read', )),
])
def test_io_hardcoded(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens if t.marker != 'import')
    assert markers == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens if t.marker != 'import')
    assert markers == expected


@pytest.mark.parametrize('text, expected', [
    ('from pathlib import Path\np = Path()\np.write_text("lol")', ('write', )),
    ('from pathlib import Path\np = Path()\np.open("w")', ('write', )),
    ('from pathlib import Path\np = Path()\nwith p.open("w"): ...', ('write', )),

    ('from pathlib import Path\np = Path()\np.open("r")', ()),          # allowed mode
    ('from pathlib import Path\np = Path()\np.open(mode="r")', ()),     # allowed mode
    ('from pathlib import Path\np = Path()\np.read_text()', ()),    # read, not write
    ('from pathlib import Path\np = Path()\np.open()', ()),         # implicit read
    ('with something.open("w"): ...', ()),                          # not pathlib
    ('something = file\nwith something.open("w"): ...', ()),        # not pathlib
    ('class Path:\n def write_text(): pass\np = Path()\np.write_text()', ()),   # not pathlib
])
def test_io_infer(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens if t.marker != 'import')
    assert markers == expected


@pytest.mark.parametrize('text, expected', [
    ('global a', ('global', )),
    ('global a, b, c', ('global', )),

    ('nonlocal a', ('global', )),
    ('nonlocal a, b, c', ('global', )),

    ('import a', ('import', )),
    ('import a as b', ('import', )),
    ('import a as b, c', ('import', )),

    ('from a import b', ('import', )),
    ('from a import b as c', ('import', )),

    ('__import__("something")', ('import', )),
])
def test_get_globals_simple(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens)
    assert markers == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens)
    assert markers == expected


def test_io_recursive_analise_body():
    text = """
    def inner(text):
        print(text)

    def outer():
        inner('hello')
    """
    text = dedent(text)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body[-1].body))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('stdout', )


def test_io_recursive_explicit_markers():
    text = """
    @deal.has('io', invalid)
    @deal.raises()
    def inner(text):
        noting()

    def outer():
        inner('hello')
    """
    text = dedent(text)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body[-1].body))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('io', )


def test_markers_from_stubs():
    text = """
    import ast

    nothing = None

    def do_nothing():
        pass

    @deal.has()
    def inner(text):
        ast.walk(None)
        not_resolvable()
        do_nothing()
        nothing()
    """
    text = dedent(text)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    stubs = StubsManager()
    tokens = list(get_markers(body=tree.body[-1].body, stubs=stubs))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('import', )



# built-in
import ast
from pathlib import Path
from textwrap import dedent

# project
from deal.linter import Checker


TEXT = """
import deal


@deal.post(lambda x: x > 0)
def test():
    return -1


@deal.raises(ValueError)
def test2():
    1 / 0
    raise ValueError
    raise KeyError
""".strip()

EXPECTED = [
    (6, 11, 'DEAL012 post contract error (-1)', Checker),
    (11, 8, 'DEAL021 raises contract error (ZeroDivisionError)', Checker),
    (13, 10, 'DEAL021 raises contract error (KeyError)', Checker),
]


def test_stdin():
    checker = Checker(tree=ast.parse(TEXT))
    errors = list(checker.run())
    assert errors == EXPECTED


def test_astroid_path(tmp_path: Path):
    path = tmp_path / 'test.py'
    path.write_text(TEXT)
    checker = Checker(tree=ast.parse(TEXT), filename=str(path))
    errors = list(checker.run())
    assert errors == EXPECTED


def test_get_funcs_invalid_syntax(tmp_path: Path):
    """
    Atom IDE flake8 plugin can call flake8 with AST with correct syntax but with path
    to code with invalid syntax. In that case, we should ignore the file and fallback
    to the passed AST.
    """
    path = tmp_path / 'test.py'
    path.write_text('1/')
    checker = Checker(tree=ast.parse(TEXT), filename=str(path))
    errors = list(checker.run())
    assert errors == EXPECTED


def test_version():
    version = Checker(tree=None, filename='stdin').version
    assert not set(version) - set('0123456789.')


def test_remove_duplicates(tmp_path):
    text = """
        import deal

        def inner():
            raise TypeError
            raise TypeError

        @deal.raises()
        def outer():
            return inner()
    """
    path = tmp_path / 'test.py'
    path.write_text(dedent(text))
    checker = Checker(tree=ast.parse(TEXT), filename=str(path))
    errors = list(checker.run())
    assert len(errors) == 1



# built-in
from typing import Callable, Type, TypeVar, overload

# app
from . import _decorators
from ._types import ExceptionType


_CallableType = TypeVar('_CallableType', bound=Callable)
_T = TypeVar('_T')


def pre(
    validator,
    *,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    """
    Decorator implementing precondition [value][value] contract.
    [Precondition][wiki] is
    a condition that must be true before the function is executed.
    Raises `PreContractError` otherwise.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :type message: str, optional
    :param exception: exception type to raise on the contract violation.
        ``PreContractError`` by default.
    :type exception: ExceptionType, optional
    :return: a function wrapper.
    :rtype: Callable[[_CallableType], _CallableType]

    ```pycon
    >>> import deal
    >>> @deal.pre(lambda a, b: a + b > 0)
    ... def example(a, b):
    ...     return (a + b) * 2
    >>> example(1, 2)
    6
    >>> example(1, -2)
    Traceback (most recent call last):
      ...
    PreContractError: expected a + b > 0 (where a=1, b=-2)

    ```

    [wiki]: https://en.wikipedia.org/wiki/Precondition
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    cls = _decorators.Pre[_CallableType]
    return cls(validator=validator, message=message, exception=exception)


def post(
    validator,
    *,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    """
    Decorator implementing postcondition [value][value] contract.
    [Postcondition][wiki] is
    a condition that must be true for the function result.
    Raises `PostContractError` otherwise.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :type message: str, optional
    :param exception: exception type to raise on the contract violation.
        ``PostContractError`` by default.
    :type exception: ExceptionType, optional
    :return: a function wrapper.
    :rtype: Callable[[_CallableType], _CallableType]

    ```pycon
    >>> import deal
    >>> @deal.post(lambda res: res > 0)
    ... def example(a, b):
    ...     return a + b
    >>> example(-1, 2)
    1
    >>> example(1, -2)
    Traceback (most recent call last):
      ...
    PostContractError: expected res > 0 (where res=-1)

    ```

    [wiki]: https://en.wikipedia.org/wiki/Postcondition
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    cls = _decorators.Post[_CallableType]
    return cls(validator=validator, message=message, exception=exception)


def ensure(
    validator,
    *,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    """
    Decorator implementing postcondition [value] contract.
    [Postcondition][wiki] is
    a condition that must be true for the function result.
    Raises `PostContractError` otherwise.
    It's like [@deal.post](#deal.post) but contract accepts as input value
    not only the function result but also the function input arguments.
    The function result is passed into validator as `result` keyword argument.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :type message: str, optional
    :param exception: exception type to raise on the contract violation.
        `PostContractError` by default.
    :type exception: ExceptionType, optional
    :return: a function wrapper.
    :rtype: Callable[[_CallableType], _CallableType]

    ```pycon
    >>> import deal
    >>> @deal.ensure(lambda a, result: a < result)
    ... def example(a):
    ...     return a * 2
    >>> example(2)
    4
    >>> example(0)
    Traceback (most recent call last):
      ...
    PostContractError: expected a < result (where result=0, a=0)

    ```

    [wiki]: https://en.wikipedia.org/wiki/Postcondition
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    cls = _decorators.Ensure[_CallableType]
    return cls(validator=validator, message=message, exception=exception)


def raises(
    *exceptions: Type[Exception],
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    """
    Decorators listing exceptions which the function can raise.
    Implements the most important [exception] contract.
    If the function raises an exception not listed in the decorator,
    `RaisesContractError` will be raised.

    :param exceptions: exceptions which the function can raise.
    :type exceptions: Type[Exception].
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :type message: str, optional
    :param exception: exception type to raise on the contract violation.
        `RaisesContractError` by default.
    :type exception: ExceptionType, optional
    :return: a function wrapper.
    :rtype: Callable[[_T], _T]

    ```pycon
    >>> import deal
    >>> @deal.raises(ZeroDivisionError, ValueError)
    ... def div(a, b):
    ...   return a / b
    ...
    >>> div(1, 0)
    Traceback (most recent call last):
        ...
    ZeroDivisionError: division by zero
    >>> div(1, '')
    Traceback (most recent call last):
        ...
     TypeError: unsupported operand type(s) for /: 'int' and 'str'
     The above exception was the direct cause of the following exception:
        ...
    RaisesContractError

    ```

    [exception] https://deal.readthedocs.io/basic/exceptions.html
    """
    cls = _decorators.Raises[_CallableType]
    return cls(*exceptions, message=message, exception=exception)


def has(
    *markers: str,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    cls = _decorators.Has[_CallableType]
    return cls(*markers, message=message, exception=exception)


def reason(
    event: Type[Exception],
    validator,
    *,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    """
    Decorator implementing [exception] contract.
    Allows to assert precondition for raised exception.
    It's like [@deal.ensure](#deal.ensure) but when instead of returning result
    the function raises an exception.

    :param event: exception raising which will trigger contract validation.
    :type event: Type[Exception].
    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :type message: str, optional
    :param exception: exception type to raise on the contract violation.
        `ReasonContractError` by default.
    :type exception: ExceptionType, optional
    :return: a function wrapper.
    :rtype: Callable[[_T], _T]

    ```pycon
    >>> import deal
    >>> @deal.reason(ZeroDivisionError, lambda a, b: b == 0)
    ... def div(a, b):
    ...   return a / (a - b)
    >>> div(2, 1)
    2.0
    >>> div(0, 0)
    Traceback (most recent call last):
        ...
    ZeroDivisionError: division by zero
    >>> div(2, 2)
    Traceback (most recent call last):
        ...
     ZeroDivisionError: division by zero
     The above exception was the direct cause of the following exception:
        ...
    ReasonContractError

    ```

    [exception]: https://deal.readthedocs.io/basic/exceptions.html
    """
    cls = _decorators.Reason[_CallableType]
    return cls(event=event, validator=validator, message=message, exception=exception)


def inv(
    validator,
    *,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_T], _T]:
    """
    Decorator implementing invariant [value][value] contract.

    [Invariant][wiki] is a condition that can be relied upon to be true during execution
    of a program. `@deal.inv` is triggered in 3 cases:

    1. Before class method execution.
    2. After class method execution.
    3. After some class attribute setting.

    Deal doesn't rollback changes on contract violation.

    :param validator: a function or validator that implements the contract.
    :param message: error message for the exception raised on contract violation.
        No error message by default.
    :type message: str, optional
    :param exception: exception type to raise on the contract violation.
        `InvContractError` by default.
    :type exception: ExceptionType, optional
    :return: a class wrapper.
    :rtype: Callable[[_T], _T]

    ```pycon
    >>> import deal
    >>> @deal.inv(lambda obj: obj.likes >= 0)
    ... class Video:
    ...   likes = 1
    ...   def like(self): self.likes += 1
    ...   def dislike(self): self.likes -= 1
    ...
    >>> v = Video()
    >>> v.dislike()
    >>> v.likes
    0
    >>> v.dislike()
    Traceback (most recent call last):
    ...
    InvContractError: expected obj.likes >= 0
    >>> v.likes
    -1
    >>> v.likes = 2
    >>> v.likes = -2
    Traceback (most recent call last):
    ...
    InvContractError: expected obj.likes >= 0
    >>> v.likes
    -2

    ```

    [wiki]: https://en.wikipedia.org/wiki/Class_invariant
    [value]: https://deal.readthedocs.io/basic/values.html
    """
    cls = _decorators.Invariant[_CallableType]
    return cls(  # type: ignore
        validator=validator,
        message=message,
        exception=exception,
    )


@overload
def safe(
    *,
    message: str = None,
    exception: ExceptionType = None,
) -> Callable[[_CallableType], _CallableType]:
    pass  # pragma: no cover


@overload
def safe(_func: _CallableType) -> _CallableType:
    pass  # pragma: no cover


def safe(_func=None, **kwargs):
    """
    Alias for [@deal.raises()](#deal.raises).
    Wraps a function that never raises an exception.

    ```pycon
    >>> import deal
    >>> @deal.safe
    ... def div(a, b):
    ...   return a / b
    ...
    >>> div(2, 4)
    0.5
    >>> div(2, 0)
    Traceback (most recent call last):
        ...
     ZeroDivisionError: division by zero
     The above exception was the direct cause of the following exception:
        ...
    RaisesContractError

    ```
    """
    if _func is None:
        return raises(**kwargs)
    return raises()(_func)


def chain(*contracts) -> Callable[[_CallableType], _CallableType]:
    """
    Decorator to chain 2 or more contracts together.
    It can be helpful to store contracts separately from the function.
    Consider using it when you have too many contracts.
    Otherwise, the function will be lost under a bunch of decorators.

    ```pycon
    >>> import deal
    >>> sum_contract = deal.chain(
    ...   deal.pre(lambda a, b: a > 0),
    ...   deal.pre(lambda a, b: b > 0),
    ...   deal.post(lambda res: res > 0),
    ... )
    >>> @sum_contract
    ... def sum(a, b):
    ...   return a + b
    ...
    >>> sum(2, 3)
    5
    >>> sum(2, -3)
    Traceback (most recent call last):
        ...
    PreContractError: expected b > 0 (where a=2, b=-3)
    >>> sum(-2, 3)
    Traceback (most recent call last):
        ...
    PreContractError: expected a > 0 (where a=-2, b=3)

    ```

    :param contracts: contracts to chain.
    :return: a function wrapper
    :rtype: Callable[[_CallableType], _CallableType]
    """
    def wrapped(func):
        for contract in contracts:
            func = contract(func)
        return func
    return wrapped


def pure(_func: _CallableType) -> _CallableType:
    """
    Decorator for [pure functions][wiki].
    Alias for `@deal.chain(deal.has(), deal.safe)`.

    Pure function has no side-effects and doesn't raise any exceptions.

    ```pycon
    >>> import deal
    >>> @deal.pure
    ... def div(a, b, log=False):
    ...   if log:
    ...     print('div called')
    ...   return a / b
    ...
    >>> div(2, 4)
    0.5
    >>> div(2, 0)
    Traceback (most recent call last):
        ...
     ZeroDivisionError: division by zero
     The above exception was the direct cause of the following exception:
        ...
    RaisesContractError
    >>> div(2, 3, log=True)
    Traceback (most recent call last):
        ...
    SilentContractError

    ```

    [wiki]: https://en.wikipedia.org/wiki/Pure_function
    """
    return chain(has(), safe)(_func)



# built-in
from typing import NoReturn, TypeVar

# external
import hypothesis
import pytest

# project
import deal


@deal.raises(ZeroDivisionError)
def div1(a: int, b: int) -> float:
    assert isinstance(a, int)
    assert isinstance(b, int)
    return a / b


@deal.pre(lambda a, b: a > 0)
@deal.pre(lambda a, b: b > 0, message='b must be positive')
def div2(a: int, b: int) -> float:
    assert isinstance(a, int)
    assert isinstance(b, int)
    assert a > 0
    assert b > 0
    return a / b


def test_count():
    for count in (1, 10, 20, 50):
        cases = deal.cases(div1, count=count)
        assert len(list(cases)) == count

        cases = deal.cases(div2, count=count)
        assert len(list(cases)) == count


def test_params_detected():
    for case in deal.cases(div1, count=10):
        assert set(case.kwargs) == {'a', 'b'}

    for case in deal.cases(div2, count=10):
        assert set(case.kwargs) == {'a', 'b'}


def test_params_type():
    for case in deal.cases(div1, count=10):
        assert type(case.kwargs['a']) is int
        assert type(case.kwargs['b']) is int

    for case in deal.cases(div2, count=10):
        assert type(case.kwargs['a']) is int
        assert type(case.kwargs['b']) is int


def test_params_ok_with_excs():
    results = []
    for case in deal.cases(div1, count=20):
        result = case()
        results.append(result)
    assert len(results) == 20
    assert any(r is NoReturn for r in results), 'no exception occured'


def test_no_bad_examples():
    for case in deal.cases(div2, count=20):
        assert case.kwargs['a'] > 0
        assert case.kwargs['b'] > 0


def test_return_type_checks():
    def div(a: int, b: int):
        return 1

    for case in deal.cases(div, count=20):
        case()

    def div(a: int, b: int) -> str:
        return 1

    with pytest.raises(TypeError):
        case = next(iter(deal.cases(div, count=20)))
        case()


def test_explicit_kwargs():
    def div(a: int, b: int):
        assert b == 4

    for case in deal.cases(div, kwargs=dict(b=4), count=20):
        case()


def test_explicit_strategy():
    def div(a: int, b: int):
        assert 0 <= b <= 4

    cases = deal.cases(
        div,
        kwargs=dict(b=hypothesis.strategies.integers(min_value=0, max_value=4)),
        count=20,
    )
    for case in cases:
        case()


def test_disable_type_checks():
    def bad(a: int) -> str:
        return a

    # type is wrong and checked
    cases = deal.cases(bad, count=1)
    case = next(cases)
    msg = 'type of the return value must be str; got int instead'
    with pytest.raises(TypeError, match=msg):
        case()

    # type is wrong and ignored
    cases = deal.cases(bad, count=1, check_types=False)
    case = next(cases)
    case()

    def good(a: int) -> int:
        return a

    # type is good
    cases = deal.cases(good, count=1)
    case = next(cases)
    case()


def test_type_var():
    T = TypeVar('T')  # noqa: N806

    def identity(a: T, b) -> T:
        return b

    kwargs = dict(kwargs={}, func=identity, exceptions=(), check_types=True)
    case = deal.TestCase(args=('ab', 'cd'), **kwargs)
    case()

    case = deal.TestCase(args=(['ab'], ['cd', 'ef']), **kwargs)
    case()

    case = deal.TestCase(args=('ab', 1), **kwargs)
    msg = 'type of the return value must be exactly str; got int instead'
    with pytest.raises(TypeError, match=msg):
        case()



# built-in
import inspect
from asyncio import iscoroutinefunction
from contextlib import suppress
from functools import update_wrapper
from typing import Any, Callable, Dict, Generic, NoReturn, TypeVar

# external
import vaa

# app
from .._exceptions import ContractError
from .._state import state
from .._types import ExceptionType


#: We use this type in many other subclasses of `Base` decorator.
_CallableType = TypeVar('_CallableType', bound=Callable)


class Base(Generic[_CallableType]):
    exception: ExceptionType = ContractError
    function: _CallableType  # pytype: disable=not-supported-yet

    def __init__(self, validator, *, message: str = None,
                 exception: ExceptionType = None):
        """
        Step 1. Set contract (validator).
        """
        self.validator = self._make_validator(validator, message=message)
        if exception:
            self.exception = exception
        if message:
            self.exception = self.exception(message)    # type: ignore

    @staticmethod
    def _make_validator(validator, message: str = None):
        if validator is None:
            return None
        # implicitly wrap in vaa all external validators
        with suppress(TypeError):
            return vaa.wrap(validator, simple=False)

        # implicitly wrap in vaa.simple only funcs with one `_` argument.
        if inspect.isfunction(validator):
            params = inspect.signature(validator).parameters
            if set(params) == {'_'}:
                return vaa.simple(validator, error=message)

        return validator

    def validate(self, *args, **kwargs) -> None:
        """
        Step 4. Process contract (validator)
        """

        if hasattr(self.validator, 'is_valid'):
            self._vaa_validation(*args, **kwargs)
        else:
            self._simple_validation(*args, **kwargs)

    def _raise(self, *, message: str = None, errors=None, params=None) -> NoReturn:
        exception = self.exception
        if isinstance(exception, Exception):
            if not message and exception.args:
                message = exception.args[0]
            exception = type(exception)

        # if errors provided, use it as error message
        if errors and isinstance(errors, str):
            message = errors
            errors = None

        # raise beautiful ContractError
        if issubclass(exception, ContractError):
            raise exception(
                message=message or '',
                validator=self.validator,
                errors=errors,
                params=params,
            )

        # raise boring custom exception
        args = []
        if message:
            args.append(message)
        if errors:
            args.append(errors)
        raise exception(*args)

    def _args_to_vars(self, *, args, kwargs: Dict[str, Any], function=None) -> Dict[str, Any]:
        """Convert args and kwargs into dict of params based on the given function.

        Function is not passed for `vaa` validators.
        For simple validators the validator is passed as function.
        If no function passed, wrapped function will be used.
        """
        keep_result = True
        if function is None:
            function = getattr(self, 'function', None)
            keep_result = False
        if function is None:
            return kwargs

        params = kwargs.copy()
        # Do not pass argument named `result` into the function.
        # It is a hack for `deal.ensure` with `vaa` validator.
        if not keep_result and 'result' in kwargs:
            kwargs = kwargs.copy()
            del kwargs['result']

        # detect original function
        function = inspect.unwrap(function)
        # assign *args to real names
        params.update(inspect.getcallargs(function, *args, **kwargs))
        return params

    def _vaa_validation(self, *args, **kwargs) -> None:
        """Validate contract using vaa wrapped validator
        """

        # if it is a decorator for a function, convert positional args into named ones.
        params = self._args_to_vars(args=args, kwargs=kwargs)

        # validate
        validator = self.validator(data=params)
        if validator.is_valid():
            return

        # if no errors returned, raise the default exception
        errors = validator.errors
        if not errors:
            self._raise(params=params)

        # Flatten single error without field to one simple str message.
        # This is for better readability of simple validators.
        if type(errors) is list:  # pragma: no cover
            if type(errors[0]) is vaa.Error:
                if len(errors) == 1:
                    if errors[0].field is None:
                        errors = errors[0].message

        self._raise(errors=errors, params=params)

    def _simple_validation(self, *args, **kwargs) -> None:
        """Validate contract using simple validator.

        Simple validator is a validator that reflects the function signature.
        """
        validation_result = self.validator(*args, **kwargs)
        # is invalid (validator returns error message)
        if isinstance(validation_result, str):
            params = self._args_to_vars(args=args, kwargs=kwargs, function=self.validator)
            self._raise(message=validation_result, params=params)
        # is valid (truely result)
        if validation_result:
            return
        # is invalid (falsy result)
        params = self._args_to_vars(args=args, kwargs=kwargs, function=self.validator)
        self._raise(params=params)

    @property
    def enabled(self) -> bool:
        return state.debug

    def __call__(self, function: _CallableType) -> _CallableType:
        """
        Step 2. Return wrapped function.
        """
        self.function = function

        def wrapped(*args, **kwargs):
            if self.enabled:
                return self.patched_function(*args, **kwargs)
            return function(*args, **kwargs)

        async def async_wrapped(*args, **kwargs):
            if self.enabled:
                return await self.async_patched_function(*args, **kwargs)
            return await function(*args, **kwargs)

        def wrapped_generator(*args, **kwargs):
            if self.enabled:
                yield from self.patched_generator(*args, **kwargs)
            else:
                yield from function(*args, **kwargs)

        if iscoroutinefunction(function):
            new_callable = update_wrapper(async_wrapped, function)
        elif inspect.isgeneratorfunction(function):
            new_callable = update_wrapper(wrapped_generator, function)
        else:
            new_callable = update_wrapper(wrapped, function)
        return new_callable  # type: ignore

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        raise NotImplementedError

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        raise NotImplementedError

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        raise NotImplementedError



# built-in
from pathlib import Path

# project
from deal._cli._lint import get_errors, lint_command


TEXT = """
import deal

@deal.post(lambda x: x > 0)
def f(x):
    return -1
"""


def test_get_errors(tmp_path: Path):
    (tmp_path / 'example.py').write_text(TEXT)
    errors = list(get_errors(paths=[tmp_path]))
    assert len(errors) == 1
    assert errors[0]['code'] == 'DEAL012'
    assert errors[0]['content'] == '    return -1'


def test_lint_command_colors(tmp_path: Path, capsys):
    (tmp_path / 'example.py').write_text(TEXT)
    count = lint_command([str(tmp_path)])
    assert count == 1

    captured = capsys.readouterr()
    assert '\x1b[34mreturn\x1b[39;49;00m -\x1b[34m1\x1b[39;49;00m' in captured.out
    assert '\x1b[95m(-1)\x1b[0m' in captured.out
    assert '\x1b[95m^\x1b[0m' in captured.out


def test_lint_command_no_color(tmp_path: Path, capsys):
    (tmp_path / 'example.py').write_text(TEXT)
    count = lint_command(['--nocolor', str(tmp_path)])
    assert count == 1

    captured = capsys.readouterr()
    exp = '6:11 DEAL012 post contract error (-1) return -1 ^'
    assert captured.out.split()[1:] == exp.split()


def test_lint_command_two_files(tmp_path: Path, capsys):
    (tmp_path / 'example1.py').write_text(TEXT)
    (tmp_path / 'example2.py').write_text(TEXT)
    count = lint_command(['--nocolor', str(tmp_path)])
    assert count == 2

    captured = capsys.readouterr()
    assert 'example1.py' in captured.out
    assert 'example2.py' in captured.out
    assert 'return -1' in captured.out
    assert '(-1)' in captured.out
    assert '^' in captured.out


def test_lint_command_two_errors(tmp_path: Path, capsys):
    (tmp_path / 'example.py').write_text('from deal import pre\n' + TEXT)
    count = lint_command(['--nocolor', str(tmp_path)])
    assert count == 2

    captured = capsys.readouterr()
    assert 'return -1' in captured.out
    assert '(-1)' in captured.out
    assert '^' in captured.out


def test_lint_command_json(tmp_path: Path, capsys):
    (tmp_path / 'example.py').write_text(TEXT)
    count = lint_command(['--json', str(tmp_path)])
    assert count == 1

    captured = capsys.readouterr()
    assert '"    return -1"' in captured.out
    assert '"-1"' in captured.out
    assert '^' not in captured.out



# built-in
import ast
from typing import Iterator, Optional

# external
import astroid

# app
from .._contract import Category
from .._stub import StubsManager
from .common import TOKENS, Extractor, Token, get_full_name, get_name, get_stub, infer
from .contracts import get_contracts
from .value import get_value


get_markers = Extractor()


@get_markers.register(*TOKENS.GLOBAL)
def handle_global(expr, **kwargs) -> Optional[Token]:
    return Token(marker='global', line=expr.lineno, col=expr.col_offset)


@get_markers.register(*TOKENS.NONLOCAL)
def handle_nonlocal(expr, **kwargs) -> Optional[Token]:
    return Token(marker='global', line=expr.lineno, col=expr.col_offset)


@get_markers.register(ast.Import)
def handle_ast_import(expr: ast.Import, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(astroid.Import)
def handle_astroid_import(expr: astroid.Import, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(ast.ImportFrom)
def handle_ast_import_from(expr: ast.ImportFrom, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(astroid.ImportFrom)
def handle_astroid_import_from(expr: astroid.ImportFrom, **kwargs) -> Optional[Token]:
    return Token(marker='import', line=expr.lineno, col=expr.col_offset)


@get_markers.register(*TOKENS.CALL)
def handle_call(expr, dive: bool = True, stubs: StubsManager = None) -> Iterator[Token]:
    token_info = dict(line=expr.lineno, col=expr.col_offset)
    name = get_name(expr.func)
    if name is None:
        return

    # stdout and stderr
    token = _check_print(expr=expr, name=name)
    if token is not None:
        yield token
        return
    if name.startswith('sys.stdout.'):
        yield Token(marker='stdout', value='sys.stdout', **token_info)
        return
    if name.startswith('sys.stderr.'):
        yield Token(marker='stderr', value='sys.stderr', **token_info)
        return

    # read and write
    if name == 'open':
        if _is_open_to_write(expr):
            yield Token(marker='write', value='open', **token_info)
        else:
            yield Token(marker='read', value='open', **token_info)
        return
    if _is_pathlib_write(expr):
        yield Token(marker='write', value='Path.open', **token_info)
        return

    # import
    if name == '__import__':
        yield Token(marker='import', **token_info)
        return

    yield from _infer_markers(expr=expr, dive=dive, stubs=stubs)


def _infer_markers(expr, dive: bool, stubs: StubsManager = None) -> Iterator[Token]:
    inferred = infer(expr=expr.func)
    stubs_found = False
    if type(expr) is astroid.Call and stubs is not None:
        for token in _markers_from_stubs(expr=expr, inferred=inferred, stubs=stubs):
            stubs_found = True
            yield token

    # Infer function call and check the function body for raises.
    # Do not dive into called function if we already found stubs for it.
    if not stubs_found and dive:
        yield from _markers_from_func(expr=expr, inferred=inferred)


@get_markers.register(*TOKENS.WITH)
def handle_with(expr, **kwargs) -> Optional[Token]:
    token_info = dict(line=expr.lineno, col=expr.col_offset)
    for item in expr.items:
        if isinstance(item, ast.withitem):
            item = item.context_expr
        else:
            item = item[0]
        if _is_pathlib_write(item):
            return Token(marker='write', value='Path.open', **token_info)
        if not isinstance(item, TOKENS.CALL):
            continue
        name = get_name(item.func)
        if name == 'open':
            if _is_open_to_write(item):
                return Token(marker='write', value='open', **token_info)
            return Token(marker='read', value='open', **token_info)
    return None


def _is_open_to_write(expr) -> bool:
    for arg in expr.args:
        if isinstance(arg, astroid.Const) and arg.value == 'w':
            return True
        if isinstance(arg, ast.Str) and 'w' in arg.s:
            return True

    if not expr.keywords:
        return False
    for arg in expr.keywords:
        if arg.arg != 'mode':
            continue
        if isinstance(arg.value, astroid.Const) and 'w' in arg.value.value:
            return True
        if isinstance(arg.value, ast.Str) and 'w' in arg.value.s:
            return True
    return False


def _is_pathlib_write(expr) -> bool:
    if not isinstance(expr, astroid.Call):
        return False
    if not isinstance(expr.func, astroid.Attribute):
        return False
    if expr.func.attrname not in ('write_text', 'write_bytes', 'open'):
        return False

    # if it's open, check that mode is "w"
    if expr.func.attrname == 'open':
        if not _is_open_to_write(expr):
            return False

    for value in infer(expr.func.expr):
        if isinstance(value, astroid.Instance):
            if value.pytype().startswith('pathlib.'):
                return True
    return False


def _markers_from_stubs(expr: astroid.Call, inferred, stubs: StubsManager) -> Iterator[Token]:
    for value in inferred:
        if type(value) is not astroid.FunctionDef:
            continue
        module_name, func_name = get_full_name(expr=value)
        stub = get_stub(module_name=module_name, expr=value, stubs=stubs)
        if stub is None:
            continue
        names = stub.get(func=func_name, contract=Category.HAS)
        for name in names:
            yield Token(marker=name, line=expr.lineno, col=expr.col_offset)


def _markers_from_func(expr, inferred) -> Iterator[Token]:
    for value in inferred:
        if type(value) is not astroid.FunctionDef:
            continue

        # recursively infer markers from the function body
        for token in get_markers(body=value.body, dive=False):
            yield Token(
                marker=token.marker,
                value=token.value,
                line=expr.lineno,
                col=expr.col_offset,
            )

        # get explicitly specified markers from `@deal.has`
        if not value.decorators:
            continue
        for category, args in get_contracts(value.decorators.nodes):
            if category != 'has':
                continue
            for arg in args:
                value = get_value(arg)
                if type(value) is not str:
                    continue
                yield Token(marker=value, line=expr.lineno, col=expr.col_offset)
    return None


def _check_print(expr, name: str) -> Optional[Token]:
    """Return token if expr is `print` function call.

    Marker type depends on `file=` keyword argument.
    If it is missed, the type is `stdout`.
    If it is `stdout` or `stderr`, the type is `stdout` or `stderr`.
    Otherwise, there is no marker. It writes something into a stream, and it's ok.
    """
    if name != 'print':
        return None
    for kwarg in (expr.keywords or []):
        if kwarg.arg != 'file':
            continue
        value = get_name(expr=kwarg.value)
        if value in ('stdout', 'sys.stdout'):
            return Token(
                marker='stdout',
                value='print',
                line=expr.lineno,
                col=expr.col_offset,
            )
        if value in ('stderr', 'sys.stderr'):
            return Token(
                marker='stderr',
                value='print',
                line=expr.lineno,
                col=expr.col_offset,
            )
        return None
    return Token(
        marker='stdout',
        value='print',
        line=expr.lineno,
        col=expr.col_offset,
    )



import sys
from pathlib import Path
from contextlib import suppress
from typing import Any, Dict, Optional, Type

import pygments
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.python import PythonLexer

from ._cached_property import cached_property
from ._source import get_validator_source
from ._state import state
from ._colors import COLORS, NOCOLORS

root = str(Path(__file__).parent)


def exception_hook(etype: Type[BaseException], value: BaseException, tb):
    """Exception hook to remove deal from the traceback for ContractError.
    """
    if not issubclass(etype, ContractError):
        return _excepthook(etype, value, tb)

    # try to reduce traceback by removing `deal` part
    prev_tb = patched_tb = tb
    while patched_tb:
        path: str = patched_tb.tb_frame.f_code.co_filename
        if path.startswith(root):
            with suppress(AttributeError):  # read-only attribute in <3.7
                prev_tb.tb_next = None
            break  # pragma: no cover
        prev_tb = patched_tb
        patched_tb = patched_tb.tb_next
    else:
        # cannot find deal in the trace, leave it as is
        patched_tb = tb

    return _excepthook(etype, value, patched_tb)


_excepthook = sys.excepthook
sys.excepthook = exception_hook


class ContractError(AssertionError):
    message: str
    errors: Optional[Any]
    validator: Any
    params: Dict[str, Any]

    def __init__(
        self,
        message: str = '',
        errors=None,
        validator=None,
        params: Dict[str, Any] = None,
    ) -> None:
        self.message = message
        self.errors = errors
        self.validator = validator
        self.params = params or {}

        args = []
        if message:
            args.append(message)
        if errors:
            args.append(errors)
        super().__init__(*args)

    @cached_property
    def source(self) -> str:
        if self.validator is None:
            return ''
        source = get_validator_source(self.validator)
        if source:
            return source
        if hasattr(self.validator, '__name__'):
            return self.validator.__name__
        return repr(self.validator)

    @cached_property
    def colored_source(self) -> str:
        source = pygments.highlight(
            code=self.source,
            lexer=PythonLexer(),
            formatter=TerminalFormatter(),
        )
        return source.rstrip()

    @cached_property
    def variables(self) -> str:
        sep = ', '
        colors = COLORS
        if not state.color:
            colors = NOCOLORS
        tmpl = '{blue}{k}{end}={magenta}{v}{end}'
        params = []
        for k, v in self.params.items():
            v = repr(v)
            if len(v) > 20:
                continue
            params.append(tmpl.format(k=k, v=v, **colors))
        return sep.join(params)

    def __str__(self) -> str:
        result = self.message
        if not result and self.errors:
            result = repr(self.errors)
        if not result and self.source:
            result = 'expected '
            if state.color:
                result += self.colored_source
            else:
                result += self.source
        if self.variables:
            result += ' (where {})'.format(self.variables)
        return result


class PreContractError(ContractError):
    pass


class PostContractError(ContractError):
    pass


class InvContractError(ContractError):
    pass


class RaisesContractError(ContractError):
    pass


class ReasonContractError(ContractError):
    pass


class MarkerError(ContractError):
    pass


class OfflineContractError(MarkerError):
    pass


class SilentContractError(MarkerError):
    pass


# Patch module name to show in repr `deal` instead of `deal._exceptions`
for cls in ContractError.__subclasses__():
    cls.__module__ = 'deal'
for cls in MarkerError.__subclasses__():
    cls.__module__ = 'deal'



# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os.path

readme = ''
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, 'README.rst')
if os.path.exists(readme_path):
    with open(readme_path, 'rb') as stream:
        readme = stream.read().decode('utf8')

setup(
    long_description=readme,
    name='deal',
    version='4.4.0',
    description='Programming by contract',
    python_requires='>=3.6',
    project_urls={"repository": "https://github.com/orsinium/deal"},
    author='Gram',
    author_email='master_fess@mail.ru',
    license='MIT',
    keywords='deal contracts pre post invariant decorators validation pythonic functional',
    classifiers=[
        'Development Status :: 5 - Production/Stable', 'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python', 'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance'
    ],
    entry_points={"flake8.extension": ["DEAL = deal.linter:Checker"]},
    packages=[
        'deal', 'deal._cli', 'deal._decorators', 'deal.linter',
        'deal.linter._extractors'
    ],
    package_dir={"": "."},
    package_data={
        "deal": ["*.typed"],
        "deal.linter": [
            "stubs/cpython/*.json", "stubs/cpython/http/*.json",
            "stubs/cpython/urllib/*.json", "stubs/dateutil/*.json",
            "stubs/dateutil/parser/*.json", "stubs/dateutil/tz/*.json",
            "stubs/dateutil/zoneinfo/*.json", "stubs/marshmallow/*.json",
            "stubs/pytz/*.json", "stubs/requests/*.json",
            "stubs/urllib3/*.json", "stubs/urllib3/contrib/*.json",
            "stubs/urllib3/util/*.json"
        ]
    },
    install_requires=[
        'astroid', 'hypothesis', 'pygments', 'typeguard', 'vaa>=0.2.1'
    ],
    extras_require={
        "dev": [
            "coverage", "isort[pyproject]", "m2r", "marshmallow", "mypy",
            "pytest", "pytest-cov", "recommonmark", "sphinx",
            "sphinx-rtd-theme", "urllib3"
        ],
        "docs": [
            "m2r", "recommonmark", "sphinx", "sphinx-rtd-theme", "urllib3"
        ],
        "tests": [
            "coverage", "isort[pyproject]", "marshmallow", "mypy", "pytest",
            "pytest-cov", "urllib3"
        ]
    },
)



from pathlib import Path
from deal.linter._rules import rules, CheckMarkers
from deal._cli._main import COMMANDS


root = Path(__file__).parent.parent / 'docs'


def test_all_codes_listed():
    path = root / 'basic' / 'linter.md'
    content = path.read_text()
    for checker in rules:
        if type(checker) is CheckMarkers:
            continue
        code = '| DEAL{:03d} |'.format(checker.code)
        assert code in content

    for marker, code in CheckMarkers.codes.items():
        code = '| DEAL{:03d} | missed marker ({})'.format(code, marker)
        assert code in content


def test_cli_included():
    path = root / 'details' / 'cli.md'
    content = path.read_text()
    for name, cmd in COMMANDS.items():
        # has header
        tmpl = '## {n}\n\n'
        line = tmpl.format(n=name)
        assert line in content

        # has autodoc
        tmpl = '```eval_rst\n.. autofunction:: deal._cli._{n}.{c}\n```'
        line = tmpl.format(n=name, c=cmd.__name__)
        assert line in content

        # header and autodoc go next to each other
        tmpl = '## {n}\n\n```eval_rst\n.. autofunction:: deal._cli._{n}.{c}\n```'
        line = tmpl.format(n=name, c=cmd.__name__)
        assert line in content


def test_examples_included():
    path = root / 'details' / 'examples.md'
    content = path.read_text()
    for path in root.parent.joinpath('examples').iterdir():
        if '__' in path.name:
            continue
        # has header
        tmpl = '## {}\n\n'
        line = tmpl.format(path.stem)
        assert line in content

        # has include
        tmpl = '```eval_rst\n.. literalinclude:: ../../examples/{}\n```'
        line = tmpl.format(path.name)
        assert line in content



from typing import Any, Iterator, NamedTuple, Optional, Set, Tuple
from trace import Trace
import ast
import sys
import inspect


class TraceResult(NamedTuple):
    file_name: str
    func_result: Any
    covered_lines: Set[int]
    all_lines: Set[int]

    @property
    def coverage(self) -> int:
        return round(len(self.covered_lines) * 100 / len(self.all_lines))


class Only:
    def __init__(self, file_name: str):
        self.file_name = file_name

    def names(self, filename: str, modulename: str) -> int:
        return int(filename != self.file_name)


def trace(func, **kwargs) -> TraceResult:
    # find the file where the original function is defined
    original_func = inspect.unwrap(func)
    file_name = original_func.__code__.co_filename

    t = Trace(trace=False)
    # Ignore everything except the file where the function is defined.
    # It makes the tracing much faster.
    t.ignore = Only(file_name)  # type: ignore

    old_trace = sys.gettrace()
    try:
        func_result: Any = t.runfunc(func, **kwargs)  # type: ignore
    finally:
        # restore previous tracer
        sys.settrace(old_trace)     # pragma: no cover
    return _collect_trace_results(  # pragma: no cover
        t=t,
        func=original_func,
        file_name=file_name,
        func_result=func_result,
    )


# this is a separate function to restore coverage for it after playing with trace hooks.
def _collect_trace_results(t: Trace, func, file_name: str, func_result) -> TraceResult:
    all_lines = _get_func_body_statements(func=func)
    first_line = min(all_lines)
    last_line = max(all_lines)

    covered_lines: Set[int] = set()
    for fname, lineno in t.counts:  # type: ignore
        assert fname == file_name
        if lineno < first_line:
            continue
        if lineno > last_line:
            continue
        covered_lines.add(lineno)

    return TraceResult(
        file_name=file_name,
        func_result=func_result,
        covered_lines=covered_lines,
        all_lines=all_lines,
    )


def _get_func_body_statements(func) -> Set[int]:
    func_name = func.__name__
    file_name = func.__code__.co_filename

    with open(file_name) as stream:
        tree = ast.parse(stream.read(), filename=file_name)

    func_node = _get_func_node(func_name=func_name, tree=tree)
    if func_node is None:
        first_line = func.__code__.co_firstlineno
        return {first_line}

    result: Set[int] = set()
    for statement in func_node.body:
        for node in ast.walk(statement):
            # skip nodes without lineno
            if not isinstance(node, ast.stmt):
                continue
            # skip docstring
            if isinstance(node, ast.Expr) and isinstance(node.value, ast.Str):
                continue
            result.add(node.lineno)
    return result


def _get_func_node(func_name: str, tree: ast.Module) -> Optional[ast.FunctionDef]:
    for node in tree.body:
        if not isinstance(node, ast.FunctionDef):
            continue
        if node.name != func_name:
            continue
        return node
    return None


def format_lines(statements: Set[int], lines: Set[int]) -> str:
    pairs = []
    for start, end in _line_ranges(statements, lines):
        if start == end:
            pairs.append(str(start))
        else:
            pairs.append('{}-{}'.format(start, end))
    return ', '.join(pairs)


def _line_ranges(statements: Set[int], lines: Set[int]) -> Iterator[Tuple[int, int]]:
    statements = sorted(statements)
    lines = sorted(lines)
    start = 0
    end = 0
    lidx = 0
    for stmt in statements:
        if lidx >= len(lines):
            break
        if stmt == lines[lidx]:
            lidx += 1
            if not start:
                start = stmt
            end = stmt
        elif start:
            yield (start, end)
            start = 0
    if start:
        yield (start, end)



# built-in
import ast
import enum
from types import MappingProxyType
from typing import Iterator

# app
from .._decorators import Has
from ._contract import Category, Contract
from ._error import Error
from ._extractors import (
    get_asserts, get_exceptions, get_imports, get_markers,
    get_pre, get_returns, get_value, has_returns,
)
from ._func import Func
from ._stub import StubsManager


rules = []


class Required(enum.Enum):
    FUNC = 'func'
    MODULE = 'module'


def register(rule):
    rules.append(rule())
    return rule


@register
class CheckImports:
    __slots__ = ()
    code = 1
    message = 'do not use `from deal import ...`, use `import deal` instead'
    required = Required.MODULE

    def __call__(self, tree: ast.Module) -> Iterator[Error]:
        for token in get_imports(tree.body):
            if token.value != 'deal':
                continue
            yield Error(
                code=self.code,
                text=self.message,
                row=token.line,
                col=token.col,
            )


@register
class CheckPre:
    __slots__ = ()
    code = 11
    message = 'pre contract error'
    required = Required.FUNC

    def __call__(self, func: Func, stubs: StubsManager = None) -> Iterator[Error]:
        # We test only contracted functions because of poor performance.
        # Inferring every called function in the whole project
        # is a really expensive operation.
        if not func.contracts:
            return
        context = func.contracts[0].context
        for token in get_pre(body=func.body, context=context):
            yield Error(
                code=self.code,
                text=token.marker or self.message,
                value=token.value,  # type: ignore
                row=token.line,
                col=token.col,
            )


@register
class CheckReturns:
    __slots__ = ()
    code = 12
    message = 'post contract error'
    required = Required.FUNC

    def __call__(self, func: Func, stubs: StubsManager = None) -> Iterator[Error]:
        for contract in func.contracts:
            if contract.category != Category.POST:
                continue
            yield from self._check(func=func, contract=contract)

    def _check(self, func: Func, contract: Contract) -> Iterator[Error]:
        for token in get_returns(body=func.body):
            try:
                result = contract.run(token.value)
            except NameError:
                # cannot resolve contract dependencies
                return

            error_info = dict(
                row=token.line,
                col=token.col,
                code=self.code,
                value=str(token.value),
            )
            if isinstance(result, str):
                yield Error(text=result, **error_info)  # type: ignore
                continue
            if not result:
                yield Error(text=self.message, **error_info)  # type: ignore


@register
class CheckRaises:
    __slots__ = ()
    code = 21
    message = 'raises contract error'
    required = Required.FUNC

    def __call__(self, func: Func, stubs: StubsManager = None) -> Iterator[Error]:
        for contract in func.contracts:
            if contract.category != Category.RAISES:
                continue
            yield from self._check(func=func, contract=contract, stubs=stubs)

    def _check(self, func: Func, contract: Contract, stubs: StubsManager = None) -> Iterator[Error]:
        allowed = contract.exceptions
        allowed_types = tuple(exc for exc in allowed if type(exc) is not str)
        for token in get_exceptions(body=func.body, stubs=stubs):
            if token.value in allowed:
                continue
            exc = token.value
            if isinstance(exc, type):
                if issubclass(exc, allowed_types):
                    continue
                exc = exc.__name__
            yield Error(
                code=self.code,
                text=self.message,
                value=str(exc),
                row=token.line,
                col=token.col,
            )


@register
class CheckAsserts:
    __slots__ = ()
    code = 31
    message = 'assert error'
    required = Required.FUNC

    def __call__(self, func: Func, stubs: StubsManager = None) -> Iterator[Error]:
        # do not validate asserts in tests
        if func.name.startswith('test_'):
            return
        for token in get_asserts(body=func.body):
            yield Error(
                code=self.code,
                text=self.message,
                value=str(token.value),
                row=token.line,
                col=token.col,
            )


@register
class CheckMarkers:
    __slots__ = ()
    code = 40
    message = 'missed marker'
    required = Required.FUNC

    codes = MappingProxyType({
        'global': 41,
        'import': 42,
        'io': 43,
        'read': 44,
        'write': 45,
        'stdout': 46,
        'stderr': 47,
        'network': 48,
    })

    def __call__(self, func: Func, stubs: StubsManager = None) -> Iterator[Error]:
        for contract in func.contracts:
            markers = None
            if contract.category == Category.HAS:
                markers = [get_value(arg) for arg in contract.args]
            elif contract.category == Category.PURE:
                markers = []
            if markers is None:
                continue
            yield from self._check(func=func, has=Has(*markers))
            return

    @classmethod
    def _check(cls, func: Func, has: Has) -> Iterator[Error]:
        # function without IO must return something
        if not has.has_io and not has_returns(body=func.body):
            yield Error(
                code=cls.codes['io'],
                text=cls.message,
                value='io',
                row=func.line,
                col=func.col,
            )

        for token in get_markers(body=func.body):
            assert token.marker
            has_marker = getattr(has, 'has_{}'.format(token.marker), None)
            if has_marker is None:
                has_marker = token.marker in has.markers
            if has_marker:
                continue
            yield Error(
                code=cls.codes.get(token.marker, 40),
                text=cls.message,
                value=token.marker,
                row=token.line,
                col=token.col,
            )



# built-in
import ast
from textwrap import dedent

# external
import astroid
import pytest

# project
from deal.linter._extractors import get_contracts


@pytest.mark.parametrize('text, expected', [
    ('@deal.post(lambda x: x>0)', ('post', )),
    ('@deal.pure', ('pure', )),
    ('@deal.pure()', ('pure', )),
    ('@deal.chain(deal.pure)', ('pure', )),
    ('@deal.chain(deal.pure, deal.post(lambda x: x>0))', ('pure', 'post')),
])
def test_get_contracts_decorators(text, expected):
    text += '\ndef f(x): pass'

    tree = astroid.parse(text)
    print(tree.repr_tree())
    decos = tree.body[-1].decorators.nodes
    returns = tuple(cat for cat, _ in get_contracts(decos))
    assert returns == expected

    tree = ast.parse(text)
    print(ast.dump(tree))
    decos = tree.body[-1].decorator_list
    returns = tuple(cat for cat, _ in get_contracts(decos))
    assert returns == expected


def test_get_contracts_infer():
    text = """
        from io import StringIO
        import deal

        contracts = deal.chain(deal.pure, deal.post(lambda x: x>0))

        @contracts
        def f(x):
            return x

        with StringIO() as contracts2:
            @contracts2
            def f(x):
                return x
    """

    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    decos = tree.body[-2].decorators.nodes
    returns = tuple(cat for cat, _ in get_contracts(decos))
    assert returns == ('pure', 'post')



import inspect
import tokenize
from textwrap import dedent
from typing import Callable, List
from vaa._internal import Simple


TokensType = List[tokenize.TokenInfo]
processors = []


def processor(func: Callable[[TokensType], TokensType]) -> Callable[[TokensType], TokensType]:
    processors.append(func)
    return func


def get_validator_source(validator) -> str:
    # get source code
    if isinstance(validator, Simple):
        validator = validator.validator
    if not hasattr(validator, '__code__'):
        return ''
    try:
        lines, _ = inspect.getsourcelines(validator.__code__)
    except OSError:
        return ''
    lines = dedent('\n'.join(lines)).split('\n')

    # tokenize
    tokens: TokensType
    try:
        tokens = _get_tokens(lines)
    except tokenize.TokenError:
        lines = _clear_lines(lines)
        tokens = _get_tokens(lines)

    # drop junk
    for processor in processors:
        tokens = processor(tokens)

    # transform back to text
    lines = tokenize.untokenize(tokens).split('\n')
    lines = _clear_lines(lines)
    if len(lines) > 1:
        return ''
    return ' '.join(lines).replace('_.', '').lstrip()


def _clear_lines(lines: List[str]) -> List[str]:
    lines = [line.rstrip() for line in lines]
    lines = [line for line in lines if line]
    # drop trailing comma
    if lines[-1] and lines[-1][-1] == ',':
        lines[-1] = lines[-1][:-1]
    return lines


def _get_tokens(lines: List[str]) -> List[tokenize.TokenInfo]:
    tokens = tokenize.generate_tokens(iter(lines).__next__)
    exclude = {tokenize.INDENT, tokenize.DEDENT, tokenize.ENDMARKER}
    return [token for token in tokens if token.type not in exclude]


@processor
def _extract_def_name(tokens: TokensType) -> TokensType:
    for token, next_token in zip(tokens, tokens[1:]):
        if token.string == 'lambda':
            return tokens
        if token.string == '@':
            return tokens
        if token.string == 'def':
            return [next_token]
        if token.string == 'class':
            return [next_token]
    return tokens


@processor
def _drop_comments(tokens: TokensType) -> TokensType:
    return [token for token in tokens if token.type != tokenize.COMMENT]


@processor
def _extract_decorator_args(tokens: TokensType) -> TokensType:
    if not tokens:
        return tokens

    # drop decorator symbol
    if tokens[0].string == '@':
        tokens = tokens[1:]

    # proceed only if is call of a deal decorator
    if tokens[0].string != 'deal' or tokens[1].string != '.':
        return tokens

    # find where decorator starts
    start = 0
    for index, token in enumerate(tokens):
        if token.string == '(':
            start = index
            break
    else:
        return tokens
    start += 1

    end = 0
    for index, token in enumerate(tokens):
        if token.string == ')':
            end = index
    return tokens[start:end]


@processor
def _extract_assignment(tokens: TokensType) -> TokensType:
    start = 0
    for index, token in enumerate(tokens):
        if token.type == tokenize.OP and '=' in token.string:
            start = index
            break
        if token.type not in (tokenize.NAME, tokenize.DOT, tokenize.NEWLINE):
            return tokens
    else:
        return tokens
    start += 1
    return tokens[start:]


@processor
def _extract_lambda(tokens: TokensType) -> TokensType:
    start = 0
    for index, (token1, token2) in enumerate(zip(tokens, tokens[1:])):
        if token1.string != '(':
            continue
        if token2.string != 'lambda':
            continue
        start = index + 1
        break
    else:
        return tokens

    end = 0
    for index, token in enumerate(tokens[start:], start=start):
        if token.string == ')':
            end = index

    return tokens[start:end]


@processor
def _extract_lambda_body(tokens: TokensType) -> TokensType:
    # find where lambda starts
    start = 0
    for index, token in enumerate(tokens):
        if token.string == 'lambda':
            start = index + 1
            break
    else:
        return tokens

    # find where lambda body starts
    for index, token in enumerate(tokens[start:], start=start):
        if token.type == tokenize.OP and ':' in token.string:
            start = index
            break
    else:
        return tokens
    start += 1

    return tokens[start:]


@processor
def _fix_line_numbers(tokens: TokensType) -> TokensType:
    if not tokens:
        return tokens
    diff = tokens[0].start[0] - 1
    new_tokens = []
    for token in tokens:
        token = token._replace(
            start=(token.start[0] - diff, token.start[1]),
            end=(token.end[0] - diff, token.end[1]),
        )
        new_tokens.append(token)
    return new_tokens



# built-in
import json
import sys
from pathlib import Path
from textwrap import dedent

# external
import astroid

# project
from deal.linter._extractors import get_exceptions
from deal.linter._stub import StubsManager


def test_stubs_in_the_root(tmp_path: Path):
    root = tmp_path / 'project'
    root.mkdir()
    (root / '__init__.py').touch()
    # (root / 'other.py').touch()
    stub = {'isnan': {'raises': ['ZeroDivisionError']}}
    (root / 'math.json').write_text(json.dumps(stub))
    stubs = StubsManager(paths=[root])

    text = """
        from math import isnan

        @deal.raises()
        def child():
            isnan()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree, stubs=stubs))
    assert returns == (ZeroDivisionError, )


def test_stubs_next_to_imported_module(tmp_path: Path):
    root = tmp_path / 'project'
    root.mkdir()
    (root / '__init__.py').touch()
    (root / 'other.py').write_text('def parent(): pass')
    stub = {'parent': {'raises': ['ZeroDivisionError', 'SomeError']}}
    (root / 'other.json').write_text(json.dumps(stub))
    stubs = StubsManager()

    text = """
        from project.other import parent

        @deal.raises()
        def child():
            parent()
    """
    sys.path.append(str(tmp_path))
    try:
        tree = astroid.parse(dedent(text))
        print(tree.repr_tree())
        func_tree = tree.body[-1].body
        returns = tuple(r.value for r in get_exceptions(body=func_tree, stubs=stubs))
        assert set(returns) == {ZeroDivisionError, 'SomeError'}
    finally:
        sys.path = sys.path[:-1]


def test_built_in_stubs():
    stubs = StubsManager()

    text = """
        from inspect import getfile

        @deal.raises()
        def child():
            getfile(1)
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree, stubs=stubs))
    assert returns == (TypeError, )


def test_no_stubs_for_module():
    stubs = StubsManager()

    text = """
        from astroid import inference_tip

        @deal.raises()
        def child():
            inference_tip()
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree, stubs=stubs))
    assert returns == ()


def test_infer_junk():
    stubs = StubsManager()

    text = """
        def another():
            return 2

        number = 3

        @deal.raises()
        def child():
            another()
            number()
            return unknown()  # uninferrable
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree, stubs=stubs))
    assert returns == ()


def test_marhsmallow_stubs():
    stubs = StubsManager()

    text = """
        from marshmallow.utils import from_iso_datetime

        @deal.raises()
        def child():
            return from_iso_datetime('example')
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    func_tree = tree.body[-1].body
    returns = tuple(r.value for r in get_exceptions(body=func_tree, stubs=stubs))
    assert returns == (ValueError,)
